import functools
import tempfile
import traceback

import numpy as np
import sklearn.datasets
import sklearn.model_selection
from numpy.linalg import LinAlgError
from sklearn import preprocessing

from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.constants import (
    BINARY_CLASSIFICATION,
    MULTICLASS_CLASSIFICATION,
    MULTILABEL_CLASSIFICATION,
    REGRESSION,
)
from autosklearn.data.xy_data_manager import XYDataManager
from autosklearn.metrics import (
    accuracy,
    balanced_accuracy,
    f1_macro,
    f1_micro,
    f1_weighted,
    log_loss,
    precision_macro,
    precision_micro,
    precision_weighted,
    recall_macro,
    recall_micro,
    recall_weighted,
)
from autosklearn.pipeline.util import get_dataset
from autosklearn.util.data import convert_to_bin

import unittest

SCORER_LIST = [
    accuracy,
    balanced_accuracy,
    f1_macro,
    f1_micro,
    f1_weighted,
    log_loss,
    precision_macro,
    precision_micro,
    precision_weighted,
    recall_macro,
    recall_micro,
    recall_weighted,
]

N_TEST_RUNS = 5


def get_evaluation_backend():
    backend_mock = unittest.mock.Mock(spec=Backend)
    backend_mock.temporary_directory = tempfile.gettempdir()

    # Assign a default data
    backend_mock.load_datamanager.return_value = (
        get_multiclass_classification_datamanager()
    )

    return backend_mock


class Dummy(object):
    def __init__(self):
        self.name = "Dummy"


class BaseEvaluatorTest(unittest.TestCase):
    def __init__(self, methodName):
        super(BaseEvaluatorTest, self).__init__(methodName)
        self.output_directories = []

    def _fit(self, evaluator):
        return self.__fit(evaluator.fit)

    def _partial_fit(self, evaluator, fold):
        partial_fit = functools.partial(evaluator.partial_fit, fold=fold)
        return self.__fit(partial_fit)

    def __fit(self, function_handle):
        """Allow us to catch known and valid exceptions for all evaluate
        scripts."""
        try:
            function_handle()
            return True
        except KeyError as e:
            if (
                "Floating-point under-/overflow occurred at epoch" in e.args[0]
                or "removed all features" in e.args[0]
                or "failed to create intent" in e.args[0]
            ):
                pass
            else:
                traceback.print_exc()
                raise e
        except ValueError as e:
            if (
                "Floating-point under-/overflow occurred at epoch" in e.args[0]
                or "removed all features" in e.args[0]
                or "failed to create intent" in e.args[0]
            ):
                pass
            else:
                raise e
        except LinAlgError as e:
            if "not positive definite, even with jitter" in e.args[0]:
                pass
            else:
                raise e
        except RuntimeWarning as e:
            if "invalid value encountered in sqrt" in e.args[0]:
                pass
            elif "divide by zero encountered in divide" in e.args[0]:
                pass
            else:
                raise e
        except UserWarning as e:
            if "FastICA did not converge" in e.args[0]:
                pass
            else:
                raise e


def get_multiclass_classification_datamanager():
    X_train, Y_train, X_test, Y_test = get_dataset("iris")
    indices = list(range(X_train.shape[0]))
    np.random.seed(1)
    np.random.shuffle(indices)
    X_train = X_train[indices]
    Y_train = Y_train[indices]
    X_test = X_test[25:]
    Y_test = Y_test[25:]

    D = Dummy()
    D.info = {"task": MULTICLASS_CLASSIFICATION, "is_sparse": False, "label_num": 3}
    D.data = {
        "X_train": X_train,
        "Y_train": Y_train,
        "X_test": X_test,
        "Y_test": Y_test,
    }
    D.feat_type = {0: "numerical", 1: "Numerical", 2: "numerical", 3: "numerical"}
    return D


def get_abalone_datamanager():
    # https://www.openml.org/d/183
    dataset_name = "abalone"
    data = sklearn.datasets.fetch_openml(data_id=183, as_frame=True)
    feat_type = {
        i: "Categorical" if x.name == "category" else "Numerical"
        for i, x in enumerate(data["data"].dtypes)
    }
    X, y = sklearn.datasets.fetch_openml(data_id=183, return_X_y=True, as_frame=False)
    y = preprocessing.LabelEncoder().fit_transform(y)
    X_train, X_test, y_train, y_test = sklearn.model_selection.train_test_split(
        X, y, random_state=1
    )

    D = XYDataManager(
        X_train,
        y_train,
        X_test,
        y_test,
        MULTICLASS_CLASSIFICATION,
        feat_type,
        dataset_name,
    )
    return D


def get_multilabel_classification_datamanager():
    X_train, Y_train, X_test, Y_test = get_dataset("iris")
    indices = list(range(X_train.shape[0]))
    np.random.seed(1)
    np.random.shuffle(indices)
    X_train = X_train[indices]
    Y_train = Y_train[indices]

    Y_train = np.array(convert_to_bin(Y_train, 3))
    Y_test = np.array(convert_to_bin(Y_test, 3))

    X_test = X_test[25:]
    Y_test = Y_test[25:]

    D = Dummy()
    D.info = {"task": MULTILABEL_CLASSIFICATION, "is_sparse": False, "label_num": 3}
    D.data = {
        "X_train": X_train,
        "Y_train": Y_train,
        "X_test": X_test,
        "Y_test": Y_test,
    }
    D.feat_type = {0: "numerical", 1: "Numerical", 2: "numerical", 3: "numerical"}
    return D


def get_binary_classification_datamanager():
    X_train, Y_train, X_test, Y_test = get_dataset("iris")
    indices = list(range(X_train.shape[0]))
    np.random.seed(1)
    np.random.shuffle(indices)
    X_train = X_train[indices]
    Y_train = Y_train[indices]

    eliminate_class_two = Y_train != 2
    X_train = X_train[eliminate_class_two]
    Y_train = Y_train[eliminate_class_two]

    eliminate_class_two = Y_test != 2
    X_test = X_test[eliminate_class_two]
    Y_test = Y_test[eliminate_class_two]

    X_test = X_test[25:]
    Y_test = Y_test[25:]

    D = Dummy()
    D.info = {"task": BINARY_CLASSIFICATION, "is_sparse": False, "label_num": 2}
    D.data = {
        "X_train": X_train,
        "Y_train": Y_train.reshape((-1, 1)),
        "X_test": X_test,
        "Y_test": Y_test.reshape((-1, 1)),
    }
    D.feat_type = {0: "numerical", 1: "Numerical", 2: "numerical", 3: "numerical"}
    return D


def get_regression_datamanager():
    X_train, Y_train, X_test, Y_test = get_dataset("boston")
    indices = list(range(X_train.shape[0]))
    np.random.seed(1)
    np.random.shuffle(indices)
    X_train = X_train[indices]
    Y_train = Y_train[indices]

    X_test = X_test[200:]
    Y_test = Y_test[200:]

    D = Dummy()
    D.info = {"task": REGRESSION, "is_sparse": False, "label_num": 1}
    D.data = {
        "X_train": X_train,
        "Y_train": Y_train.reshape((-1, 1)),
        "X_test": X_test,
        "Y_test": Y_test.reshape((-1, 1)),
    }
    D.feat_type = {i: "numerical" for i in range(X_train.shape[1])}
    return D


def get_500_classes_datamanager():
    weights = ([0.002] * 475) + ([0.001] * 25)
    X, Y = sklearn.datasets.make_classification(
        n_samples=1000,
        n_features=20,
        n_classes=500,
        n_clusters_per_class=1,
        n_informative=15,
        n_redundant=5,
        n_repeated=0,
        weights=weights,
        flip_y=0,
        class_sep=1.0,
        hypercube=True,
        shift=None,
        scale=1.0,
        shuffle=True,
        random_state=1,
    )

    D = Dummy()
    D.info = {"task": MULTICLASS_CLASSIFICATION, "is_sparse": False, "label_num": 500}
    D.data = {
        "X_train": X[:700],
        "Y_train": Y[:700],
        "X_test": X[710:],
        "Y_test": Y[710:],
    }
    D.feat_type = {i: "numerical" for i in range(20)}
    return D


def get_dataset_getters():
    return [
        get_binary_classification_datamanager,
        get_multiclass_classification_datamanager,
        get_multilabel_classification_datamanager,
        get_500_classes_datamanager,
        get_abalone_datamanager,
        get_regression_datamanager,
    ]



from typing import Optional, Union

import copy
from itertools import product

import numpy as np
from ConfigSpace.configuration_space import Configuration, ConfigurationSpace
from ConfigSpace.forbidden import ForbiddenAndConjunction, ForbiddenEqualsClause
from sklearn.base import RegressorMixin

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import BasePipeline
from autosklearn.pipeline.components import (
    feature_preprocessing as feature_preprocessing_components,
)
from autosklearn.pipeline.components import regression as regression_components
from autosklearn.pipeline.components.data_preprocessing import DataPreprocessorChoice
from autosklearn.pipeline.constants import SPARSE


class SimpleRegressionPipeline(RegressorMixin, BasePipeline):
    """This class implements the regression task.

    It implements a pipeline, which includes one preprocessing step and one
    regression algorithm. It can render a search space including all known
    regression and preprocessing algorithms.

    Contrary to the sklearn API it is not possible to enumerate the
    possible parameters in the __init__ function because we only know the
    available regressors at runtime. For this reason the user must
    specifiy the parameters by passing an instance of
    ConfigSpace.configuration_space.Configuration.

    Parameters
    ----------
    config : ConfigSpace.configuration_space.Configuration
        The configuration to evaluate.

    random_state : Optional[int | RandomState]
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance
        used by `np.random`.

    Attributes
    ----------
    _estimator : The underlying scikit-learn regression model. This
        variable is assigned after a call to the
        :meth:`autosklearn.pipeline.regression.SimpleRegressionPipeline.fit`
        method.

    _preprocessor : The underlying scikit-learn preprocessing algorithm. This
        variable is only assigned if a preprocessor is specified and
        after a call to the
        :meth:`autosklearn.pipeline.regression.SimpleRegressionPipeline.fit`
        method.

    See also
    --------

    References
    ----------

    Examples
    --------

    """

    def __init__(
        self,
        config: Optional[Configuration] = None,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        steps=None,
        dataset_properties=None,
        include=None,
        exclude=None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        init_params=None,
    ):
        self._output_dtype = np.float32
        if dataset_properties is None:
            dataset_properties = dict()
        if "target_type" not in dataset_properties:
            dataset_properties["target_type"] = "regression"
        super().__init__(
            feat_type=feat_type,
            config=config,
            steps=steps,
            dataset_properties=dataset_properties,
            include=include,
            exclude=exclude,
            random_state=random_state,
            init_params=init_params,
        )

    def fit_estimator(self, X, y, **fit_params):
        self.y_max_ = np.nanmax(y)
        self.y_min_ = np.nanmin(y)
        return super(SimpleRegressionPipeline, self).fit_estimator(X, y, **fit_params)

    def iterative_fit(self, X, y, n_iter=1, **fit_params):
        self.y_max_ = np.nanmax(y)
        self.y_min_ = np.nanmin(y)
        return super(SimpleRegressionPipeline, self).iterative_fit(
            X, y, n_iter=n_iter, **fit_params
        )

    def predict(self, X, batch_size=None):
        y = super().predict(X, batch_size=batch_size)
        y[y > (2 * self.y_max_)] = 2 * self.y_max_
        if self.y_min_ < 0:
            y[y < (2 * self.y_min_)] = 2 * self.y_min_
        elif self.y_min_ > 0:
            y[y < (0.5 * self.y_min_)] = 0.5 * self.y_min_
        return y

    def _get_hyperparameter_search_space(
        self,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        include=None,
        exclude=None,
        dataset_properties=None,
    ):
        """Return the configuration space for the CASH problem.

        Parameters
        ----------
        include : dict
            If include is given, only the modules specified for nodes
            are used. Specify them by their module name; e.g., to include
            only the SVM use :python:`include={'regressor':['svr']}`.

        exclude : dict
            If exclude is given, only the components specified for nodes
            are used. Specify them by their module name; e.g., to include
            all regressors except the SVM use
            :python:`exclude=['regressor': 'svr']`.

        Returns
        -------
        cs : ConfigSpace.configuration_space.Configuration
            The configuration space describing the SimpleRegressionClassifier.
        """
        cs = ConfigurationSpace()

        if dataset_properties is None or not isinstance(dataset_properties, dict):
            dataset_properties = dict()
        if "target_type" not in dataset_properties:
            dataset_properties["target_type"] = "regression"
        if dataset_properties["target_type"] != "regression":
            dataset_properties["target_type"] = "regression"

        if "sparse" not in dataset_properties:
            # This dataset is probably dense
            dataset_properties["sparse"] = False

        cs = self._get_base_search_space(
            cs=cs,
            feat_type=feat_type,
            dataset_properties=dataset_properties,
            exclude=exclude,
            include=include,
            pipeline=self.steps,
        )

        regressors = cs.get_hyperparameter("regressor:__choice__").choices
        preprocessors = cs.get_hyperparameter("feature_preprocessor:__choice__").choices
        available_regressors = self._final_estimator.get_available_components(
            dataset_properties
        )

        possible_default_regressor = copy.copy(list(available_regressors.keys()))
        default = cs.get_hyperparameter("regressor:__choice__").default_value
        del possible_default_regressor[possible_default_regressor.index(default)]

        # A regressor which can handle sparse data after the densifier is
        # forbidden for memory issues
        for key in regressors:
            if (
                SPARSE
                in available_regressors[key].get_properties(dataset_properties=None)[
                    "input"
                ]
            ):
                if "densifier" in preprocessors:
                    while True:
                        try:
                            forb_reg = ForbiddenEqualsClause(
                                cs.get_hyperparameter("regressor:__choice__"), key
                            )
                            forb_fpp = ForbiddenEqualsClause(
                                cs.get_hyperparameter(
                                    "feature_preprocessor:__choice__"
                                ),
                                "densifier",
                            )
                            cs.add_forbidden_clause(
                                ForbiddenAndConjunction(forb_reg, forb_fpp)
                            )
                            # Success
                            break
                        except ValueError:
                            # Change the default and try again
                            try:
                                default = possible_default_regressor.pop()
                            except IndexError:
                                raise ValueError(
                                    "Cannot find a legal default configuration."
                                )
                            cs.get_hyperparameter(
                                "regressor:__choice__"
                            ).default_value = default

        # which would take too long
        # Combinations of tree-based models with feature learning:
        regressors_ = [
            "adaboost",
            "ard_regression",
            "decision_tree",
            "extra_trees",
            "gaussian_process",
            "gradient_boosting",
            "k_nearest_neighbors",
            "libsvm_svr",
            "mlp",
            "random_forest",
        ]
        feature_learning_ = ["kitchen_sinks", "kernel_pca", "nystroem_sampler"]

        for r, f in product(regressors_, feature_learning_):
            if r not in regressors:
                continue
            if f not in preprocessors:
                continue
            while True:
                try:
                    cs.add_forbidden_clause(
                        ForbiddenAndConjunction(
                            ForbiddenEqualsClause(
                                cs.get_hyperparameter("regressor:__choice__"), r
                            ),
                            ForbiddenEqualsClause(
                                cs.get_hyperparameter(
                                    "feature_preprocessor:__choice__"
                                ),
                                f,
                            ),
                        )
                    )
                    break
                except KeyError:
                    break
                except ValueError:
                    # Change the default and try again
                    try:
                        default = possible_default_regressor.pop()
                    except IndexError:
                        raise ValueError("Cannot find a legal default configuration.")
                    cs.get_hyperparameter(
                        "regressor:__choice__"
                    ).default_value = default

        self.configuration_space = cs
        self.dataset_properties = dataset_properties
        return cs

    def _get_estimator_components(self):
        return regression_components._regressors

    def _get_pipeline_steps(
        self, dataset_properties, feat_type: Optional[FEAT_TYPE_TYPE] = None
    ):
        steps = []

        default_dataset_properties = {"target_type": "regression"}
        if dataset_properties is not None and isinstance(dataset_properties, dict):
            default_dataset_properties.update(dataset_properties)

        steps.extend(
            [
                [
                    "data_preprocessor",
                    DataPreprocessorChoice(
                        feat_type=feat_type,
                        dataset_properties=default_dataset_properties,
                        random_state=self.random_state,
                    ),
                ],
                [
                    "feature_preprocessor",
                    feature_preprocessing_components.FeaturePreprocessorChoice(
                        feat_type=feat_type,
                        dataset_properties=default_dataset_properties,
                        random_state=self.random_state,
                    ),
                ],
                [
                    "regressor",
                    regression_components.RegressorChoice(
                        feat_type=feat_type,
                        dataset_properties=default_dataset_properties,
                        random_state=self.random_state,
                    ),
                ],
            ]
        )

        return steps

    def _get_estimator_hyperparameter_name(self):
        return "regressor"



import numpy as np
from scipy import sparse

from autosklearn.pipeline.components.data_preprocessing.categorical_encoding.no_encoding import (  # noqa: E501
    NoEncoding,
)
from autosklearn.pipeline.components.data_preprocessing.categorical_encoding.one_hot_encoding import (  # noqa: E501
    OneHotEncoder,
)
from autosklearn.pipeline.util import _test_preprocessing

import unittest


def create_X(instances=1000, n_feats=10, categs_per_feat=5, seed=0):
    rs = np.random.RandomState(seed)
    size = (instances, n_feats)
    X = rs.randint(0, categs_per_feat, size=size)
    return X


class OneHotEncoderTest(unittest.TestCase):
    def setUp(self):
        self.X_train = create_X()

    def test_data_type_consistency(self):
        X = np.random.randint(3, 6, (3, 4))
        Y = OneHotEncoder().fit_transform(X)
        self.assertFalse(sparse.issparse(Y))

        X = sparse.csc_matrix(
            ([3, 6, 4, 5], ([0, 1, 2, 1], [3, 2, 1, 0])), shape=(3, 4)
        )
        Y = OneHotEncoder().fit_transform(X)
        self.assertTrue(sparse.issparse(Y))

    def test_default_configuration(self):
        transformations = []
        for i in range(2):
            configuration_space = OneHotEncoder.get_hyperparameter_search_space()
            default_config = configuration_space.get_default_configuration()

            preprocessor = OneHotEncoder(random_state=1, **default_config)

            transformer = preprocessor.fit(self.X_train.copy())
            Xt = transformer.transform(self.X_train.copy())
            transformations.append(Xt)
            if len(transformations) > 1:
                np.testing.assert_array_equal(transformations[-1], transformations[-2])

    def test_default_configuration_no_encoding(self):
        transformations = []
        for i in range(2):
            transformation, original = _test_preprocessing(NoEncoding)
            self.assertEqual(transformation.shape, original.shape)
            self.assertTrue((transformation == original).all())
            transformations.append(transformation)
            if len(transformations) > 1:
                self.assertTrue((transformations[-1] == transformations[-2]).all())

    def test_default_configuration_sparse_data(self):
        transformations = []

        self.X_train[~np.isfinite(self.X_train)] = 0
        self.X_train = sparse.csc_matrix(self.X_train)

        for i in range(2):
            configuration_space = OneHotEncoder.get_hyperparameter_search_space()
            default_config = configuration_space.get_default_configuration()

            preprocessor = OneHotEncoder(random_state=1, **default_config)

            transformer = preprocessor.fit(self.X_train.copy())
            Xt = transformer.transform(self.X_train.copy())
            transformations.append(Xt)
            if len(transformations) > 1:
                self.assertEqual(
                    (transformations[-1] != transformations[-2]).count_nonzero(), 0
                )

    def test_default_configuration_sparse_no_encoding(self):
        transformations = []

        for i in range(2):
            transformation, original = _test_preprocessing(NoEncoding, make_sparse=True)
            self.assertEqual(transformation.shape, original.shape)
            self.assertTrue((transformation.todense() == original.todense()).all())
            transformations.append(transformation)
            if len(transformations) > 1:
                self.assertEqual(
                    (transformations[-1] != transformations[-2]).count_nonzero(), 0
                )



from __future__ import annotations

from typing import Dict, List, Sequence, Tuple, Union

import random
import warnings
from collections import Counter

import numpy as np
from sklearn.utils import check_random_state

from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.constants import TASK_TYPES
from autosklearn.data.validation import SUPPORTED_FEAT_TYPES
from autosklearn.ensemble_building.run import Run
from autosklearn.ensembles.abstract_ensemble import AbstractEnsemble
from autosklearn.metrics import Scorer, calculate_losses
from autosklearn.pipeline.base import BasePipeline


class EnsembleSelection(AbstractEnsemble):
    def __init__(
        self,
        task_type: int,
        metrics: Sequence[Scorer] | Scorer,
        backend: Backend,
        ensemble_size: int = 50,
        bagging: bool = False,
        mode: str = "fast",
        random_state: int | np.random.RandomState | None = None,
    ) -> None:
        """An ensemble of selected algorithms

        Fitting an EnsembleSelection generates an ensemble from the the models
        generated during the search process. Can be further used for prediction.

        Parameters
        ----------
        task_type: int
            An identifier indicating which task is being performed.

        metrics: Sequence[Scorer] | Scorer
            The metric used to evaluate the models. If multiple metrics are passed,
            ensemble selection only optimizes for the first

        backend : Backend
            Gives access to the backend of Auto-sklearn. Not used by Ensemble Selection.

        bagging: bool = False
            Whether to use bagging in ensemble selection

        mode: str in ['fast', 'slow'] = 'fast'
            Which kind of ensemble generation to use
            * 'slow' - The original method used in Rich Caruana's ensemble selection.
            * 'fast' - A faster version of Rich Caruanas' ensemble selection.

        random_state: int | RandomState | None = None
            The random_state used for ensemble selection.

            * None - Uses numpy's default RandomState object
            * int - Successive calls to fit will produce the same results
            * RandomState - Truly random, each call to fit will produce
              different results, even with the same object.

        References
        ----------
        | Ensemble selection from libraries of models
        | Rich Caruana, Alexandru Niculescu-Mizil, Geoff Crew and Alex Ksikes
        | ICML 2004
        | https://dl.acm.org/doi/10.1145/1015330.1015432
        | https://www.cs.cornell.edu/~caruana/ctp/ct.papers/caruana.icml04.icdm06long.pdf
        """  # noqa: E501
        self.ensemble_size = ensemble_size
        self.task_type = task_type
        if isinstance(metrics, Sequence):
            if len(metrics) > 1:
                warnings.warn(
                    "Ensemble selection can only optimize one metric, "
                    "but multiple metrics were passed, dropping all "
                    "except for the first metric."
                )
            self.metric = metrics[0]
        else:
            self.metric = metrics
        self.bagging = bagging
        self.mode = mode

        # Behaviour similar to sklearn
        #   int - Deteriministic with succesive calls to fit
        #   RandomState - Successive calls to fit will produce differences
        #   None - Uses numpmys global singleton RandomState
        # https://scikit-learn.org/stable/common_pitfalls.html#controlling-randomness
        self.random_state = random_state

    def fit(
        self,
        base_models_predictions: List[np.ndarray],
        true_targets: np.ndarray,
        model_identifiers: List[Tuple[int, int, float]],
        runs: Sequence[Run],
        X_data: SUPPORTED_FEAT_TYPES | None = None,
    ) -> EnsembleSelection:
        self.ensemble_size = int(self.ensemble_size)
        if self.ensemble_size < 1:
            raise ValueError("Ensemble size cannot be less than one!")
        if self.task_type not in TASK_TYPES:
            raise ValueError("Unknown task type %s." % self.task_type)
        if not isinstance(self.metric, Scorer):
            raise ValueError(
                "The provided metric must be an instance of Scorer, "
                "nevertheless it is {}({})".format(
                    self.metric,
                    type(self.metric),
                )
            )
        if self.mode not in ("fast", "slow"):
            raise ValueError("Unknown mode %s" % self.mode)

        if self.bagging:
            self._bagging(base_models_predictions, true_targets)
        else:
            self._fit(
                predictions=base_models_predictions,
                X_data=X_data,
                labels=true_targets,
            )
        self._calculate_weights()
        self.identifiers_ = model_identifiers
        return self

    def _fit(
        self,
        predictions: List[np.ndarray],
        labels: np.ndarray,
        *,
        X_data: SUPPORTED_FEAT_TYPES | None = None,
    ) -> EnsembleSelection:
        if self.mode == "fast":
            self._fast(predictions=predictions, X_data=X_data, labels=labels)
        else:
            self._slow(predictions=predictions, X_data=X_data, labels=labels)
        return self

    def _fast(
        self,
        predictions: List[np.ndarray],
        labels: np.ndarray,
        *,
        X_data: SUPPORTED_FEAT_TYPES | None = None,
    ) -> None:
        """Fast version of Rich Caruana's ensemble selection method."""
        self.num_input_models_ = len(predictions)
        rand = check_random_state(self.random_state)

        ensemble = []  # type: List[np.ndarray]
        trajectory = []
        order = []

        ensemble_size = self.ensemble_size

        weighted_ensemble_prediction = np.zeros(
            predictions[0].shape,
            dtype=np.float64,
        )
        fant_ensemble_prediction = np.zeros(
            weighted_ensemble_prediction.shape,
            dtype=np.float64,
        )
        for i in range(ensemble_size):
            losses = np.zeros(
                (len(predictions)),
                dtype=np.float64,
            )
            s = len(ensemble)
            if s > 0:
                np.add(
                    weighted_ensemble_prediction,
                    ensemble[-1],
                    out=weighted_ensemble_prediction,
                )

            # Memory-efficient averaging!
            for j, pred in enumerate(predictions):
                # fant_ensemble_prediction is the prediction of the current ensemble
                # and should be
                #
                #   ([predictions[selected_prev_iterations] + predictions[j])/(s+1)
                #
                # We overwrite the contents of fant_ensemble_prediction directly with
                # weighted_ensemble_prediction + new_prediction and then scale for avg
                np.add(weighted_ensemble_prediction, pred, out=fant_ensemble_prediction)
                np.multiply(
                    fant_ensemble_prediction,
                    (1.0 / float(s + 1)),
                    out=fant_ensemble_prediction,
                )

                losses[j] = calculate_losses(
                    solution=labels,
                    prediction=fant_ensemble_prediction,
                    task_type=self.task_type,
                    metrics=[self.metric],
                    X_data=X_data,
                    scoring_functions=None,
                )[self.metric.name]

            all_best = np.argwhere(losses == np.nanmin(losses)).flatten()

            best = rand.choice(all_best)

            ensemble.append(predictions[best])
            trajectory.append(losses[best])
            order.append(best)

            # Handle special case
            if len(predictions) == 1:
                break

        self.indices_ = order
        self.trajectory_ = trajectory
        self.train_loss_ = trajectory[-1]

    def _slow(
        self,
        predictions: List[np.ndarray],
        labels: np.ndarray,
        *,
        X_data: SUPPORTED_FEAT_TYPES | None = None,
    ) -> None:
        """Rich Caruana's ensemble selection method."""
        self.num_input_models_ = len(predictions)

        ensemble = []
        trajectory = []
        order = []

        ensemble_size = self.ensemble_size

        for i in range(ensemble_size):
            losses = np.zeros(
                [np.shape(predictions)[0]],
                dtype=np.float64,
            )
            for j, pred in enumerate(predictions):
                ensemble.append(pred)
                ensemble_prediction = np.mean(np.array(ensemble), axis=0)
                losses[j] = calculate_losses(
                    solution=labels,
                    prediction=ensemble_prediction,
                    task_type=self.task_type,
                    metrics=[self.metric],
                    X_data=X_data,
                    scoring_functions=None,
                )[self.metric.name]
                ensemble.pop()
            best = np.nanargmin(losses)
            ensemble.append(predictions[best])
            trajectory.append(losses[best])
            order.append(best)

            # Handle special case
            if len(predictions) == 1:
                break

        self.indices_ = np.array(
            order,
            dtype=np.int64,
        )
        self.trajectory_ = np.array(
            trajectory,
            dtype=np.float64,
        )
        self.train_loss_ = trajectory[-1]

    def _calculate_weights(self) -> None:
        ensemble_members = Counter(self.indices_).most_common()
        weights = np.zeros(
            (self.num_input_models_,),
            dtype=np.float64,
        )
        for ensemble_member in ensemble_members:
            weight = float(ensemble_member[1]) / self.ensemble_size
            weights[ensemble_member[0]] = weight

        if np.sum(weights) < 1:
            weights = weights / np.sum(weights)

        self.weights_ = weights

    def _bagging(
        self,
        predictions: List[np.ndarray],
        labels: np.ndarray,
        fraction: float = 0.5,
        n_bags: int = 20,
    ) -> np.ndarray:
        """Rich Caruana's ensemble selection method with bagging."""
        raise ValueError("Bagging might not work with class-based interface!")
        n_models = predictions.shape[0]
        bag_size = int(n_models * fraction)

        order_of_each_bag = []
        for j in range(n_bags):
            # Bagging a set of models
            indices = sorted(random.sample(range(0, n_models), bag_size))
            bag = predictions[indices, :, :]
            order, _ = self._fit(predictions=bag, labels=labels)
            order_of_each_bag.append(order)

        return np.array(
            order_of_each_bag,
            dtype=np.int64,
        )

    def predict(
        self, base_models_predictions: Union[np.ndarray, List[np.ndarray]]
    ) -> np.ndarray:

        average = np.zeros_like(base_models_predictions[0], dtype=np.float64)
        tmp_predictions = np.empty_like(base_models_predictions[0], dtype=np.float64)

        # if predictions.shape[0] == len(self.weights_),
        # predictions include those of zero-weight models.
        if len(base_models_predictions) == len(self.weights_):
            for pred, weight in zip(base_models_predictions, self.weights_):
                np.multiply(pred, weight, out=tmp_predictions)
                np.add(average, tmp_predictions, out=average)

        # if prediction model.shape[0] == len(non_null_weights),
        # predictions do not include those of zero-weight models.
        elif len(base_models_predictions) == np.count_nonzero(self.weights_):
            non_null_weights = [w for w in self.weights_ if w > 0]
            for pred, weight in zip(base_models_predictions, non_null_weights):
                np.multiply(pred, weight, out=tmp_predictions)
                np.add(average, tmp_predictions, out=average)

        # If none of the above applies, then something must have gone wrong.
        else:
            raise ValueError(
                "The dimensions of ensemble predictions"
                " and ensemble weights do not match!"
            )
        del tmp_predictions
        return average

    def __str__(self) -> str:
        trajectory_str = " ".join(
            [f"{id}: {perf:.5f}" for id, perf in enumerate(self.trajectory_)]
        )
        identifiers_str = " ".join(
            [
                f"{identifier}"
                for idx, identifier in enumerate(self.identifiers_)
                if self.weights_[idx] > 0
            ]
        )
        return (
            "Ensemble Selection:\n"
            f"\tTrajectory: {trajectory_str}\n"
            f"\tMembers: {self.indices_}\n"
            f"\tWeights: {self.weights_}\n"
            f"\tIdentifiers: {identifiers_str}\n"
        )

    def get_models_with_weights(
        self, models: Dict[Tuple[int, int, float], BasePipeline]
    ) -> List[Tuple[float, BasePipeline]]:
        output = []
        for i, weight in enumerate(self.weights_):
            if weight > 0.0:
                identifier = self.identifiers_[i]
                model = models[identifier]
                output.append((weight, model))

        output.sort(reverse=True, key=lambda t: t[0])

        return output

    def get_identifiers_with_weights(
        self,
    ) -> List[Tuple[Tuple[int, int, float], float]]:
        return list(zip(self.identifiers_, self.weights_))

    def get_selected_model_identifiers(self) -> List[Tuple[int, int, float]]:
        output = []

        for i, weight in enumerate(self.weights_):
            identifier = self.identifiers_[i]
            if weight > 0.0:
                output.append(identifier)

        return output

    def get_validation_performance(self) -> float:
        return self.trajectory_[-1]



from abc import ABCMeta, abstractmethod

import time
from io import StringIO

import arff
import scipy.sparse


class AbstractMetaFeature(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def _calculate(cls, X, y, logger, feat_type):
        pass

    def __call__(self, X, y, logger, feat_type=None):
        if feat_type is None:
            feat_type = {i: "numerical" for i in range(X.shape[1])}
        starttime = time.time()
        try:
            if scipy.sparse.issparse(X) and hasattr(self, "_calculate_sparse"):
                value = self._calculate_sparse(X, y, logger, feat_type)
            else:
                value = self._calculate(X, y, logger, feat_type)
            comment = ""
        except MemoryError:
            value = None
            comment = "Memory Error"

        endtime = time.time()
        return MetaFeatureValue(
            self.__class__.__name__,
            self.type_,
            0,
            0,
            value,
            endtime - starttime,
            comment=comment,
        )


class MetaFeature(AbstractMetaFeature):
    def __init__(self):
        super(MetaFeature, self).__init__()
        self.type_ = "METAFEATURE"


class HelperFunction(AbstractMetaFeature):
    def __init__(self):
        super(HelperFunction, self).__init__()
        self.type_ = "HELPERFUNCTION"


class MetaFeatureValue(object):
    def __init__(self, name, type_, fold, repeat, value, time, comment=""):
        self.name = name
        self.type_ = type_
        self.fold = fold
        self.repeat = repeat
        self.value = value
        self.time = time
        self.comment = comment

    def to_arff_row(self):
        if self.type_ == "METAFEATURE":
            value = self.value
        else:
            value = "?"

        return [
            self.name,
            self.type_,
            self.fold,
            self.repeat,
            value,
            self.time,
            self.comment,
        ]

    def __repr__(self):
        repr = (
            "%s (type: %s, fold: %d, repeat: %d, value: %s, time: %3.3f, "
            "comment: %s)"
        )
        repr = repr % tuple(
            self.to_arff_row()[:4]
            + [str(self.to_arff_row()[4])]
            + self.to_arff_row()[5:]
        )
        return repr


class DatasetMetafeatures(object):
    def __init__(self, dataset_name, metafeature_values):
        self.dataset_name = dataset_name
        self.metafeature_values = metafeature_values

    def _get_arff(self):
        output = dict()
        output["relation"] = "metafeatures_%s" % (self.dataset_name)
        output["description"] = ""
        output["attributes"] = [
            ("name", "STRING"),
            ("type", "STRING"),
            ("fold", "NUMERIC"),
            ("repeat", "NUMERIC"),
            ("value", "NUMERIC"),
            ("time", "NUMERIC"),
            ("comment", "STRING"),
        ]
        output["data"] = []

        for key in sorted(self.metafeature_values):
            output["data"].append(self.metafeature_values[key].to_arff_row())
        return output

    def dumps(self):
        return self._get_arff()

    def dump(self, path_or_filehandle):
        output = self._get_arff()

        if isinstance(path_or_filehandle, str):
            with open(path_or_filehandle, "w") as fh:
                arff.dump(output, fh)
        else:
            arff.dump(output, path_or_filehandle)

    @classmethod
    def load(cls, path_or_filehandle):

        if isinstance(path_or_filehandle, str):
            with open(path_or_filehandle) as fh:
                input = arff.load(fh)
        else:
            input = arff.load(path_or_filehandle)

        dataset_name = input["relation"].replace("metafeatures_", "")
        metafeature_values = []
        for item in input["data"]:
            mf = MetaFeatureValue(*item)
            metafeature_values.append(mf)

        return cls(dataset_name, metafeature_values)

    def __repr__(self, verbosity=0):
        repr = StringIO()
        repr.write("Metafeatures for dataset %s\n" % self.dataset_name)
        for name in self.metafeature_values:
            if verbosity == 0 and self.metafeature_values[name].type_ != "METAFEATURE":
                continue
            if verbosity == 0:
                repr.write(
                    "  %s: %s\n" % (str(name), str(self.metafeature_values[name].value))
                )
            elif verbosity >= 1:
                repr.write(
                    "  %s: %10s  (%10fs)\n"
                    % (
                        str(name),
                        str(self.metafeature_values[name].value)[:10],
                        self.metafeature_values[name].time,
                    )
                )

            # Add the reason for a crash if one happened!
            if verbosity > 1 and self.metafeature_values[name].comment:
                repr.write("    %s\n" % self.metafeature_values[name].comment)

        return repr.getvalue()

    def keys(self):
        return self.metafeature_values.keys()

    def __getitem__(self, item):
        return self.metafeature_values[item]



import numpy as np
import pandas as pd
import sklearn.utils
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import MinMaxScaler


class KNearestDatasets(object):
    def __init__(self, logger, metric="l1", random_state=None, metric_params=None):
        self.logger = logger

        self.metric = metric
        self.model = None
        self.metric_params = metric_params
        self.metafeatures = None
        self.runs = None
        self.best_configuration_per_dataset = None
        self.random_state = sklearn.utils.check_random_state(random_state)
        self.scaler = MinMaxScaler()

        if self.metric_params is None:
            self.metric_params = {}

    def fit(self, metafeatures, runs):
        """Fit the Nearest Neighbor model.

        Parameters
        ----------
        metafeatures : pandas.DataFrame
            A pandas dataframe. Each row represents a dataset, each column a
            metafeature.
        runs : dict
            Dictionary containing a list of runs for each dataset.
        """
        assert isinstance(metafeatures, pd.DataFrame)
        assert metafeatures.values.dtype in (np.float32, np.float64)
        assert np.isfinite(metafeatures.values).all()
        assert isinstance(runs, pd.DataFrame)
        assert runs.shape[1] == metafeatures.shape[0], (
            runs.shape[1],
            metafeatures.shape[0],
        )

        self.metafeatures = metafeatures
        self.runs = runs
        self.num_datasets = runs.shape[1]

        # Fit the metafeatures for scaler
        self.scaler.fit(self.metafeatures)

        # for each dataset, sort the runs according to their result
        best_configuration_per_dataset = {}
        for dataset_name in runs:
            if not np.isfinite(runs[dataset_name]).any():
                best_configuration_per_dataset[dataset_name] = None
            else:
                configuration_idx = runs[dataset_name].index[
                    np.nanargmin(runs[dataset_name].values)
                ]
                best_configuration_per_dataset[dataset_name] = configuration_idx

        self.best_configuration_per_dataset = best_configuration_per_dataset

        if callable(self.metric):
            self._metric = self.metric
            self._p = 0
        elif self.metric.lower() == "l1":
            self._metric = "minkowski"
            self._p = 1
        elif self.metric.lower() == "l2":
            self._metric = "minkowski"
            self._p = 2
        else:
            raise ValueError(self.metric)

        self._nearest_neighbors = NearestNeighbors(
            n_neighbors=self.num_datasets,
            radius=None,
            algorithm="brute",
            leaf_size=30,
            metric=self._metric,
            p=self._p,
            metric_params=self.metric_params,
        )

    def kNearestDatasets(self, x, k=1, return_distance=False):
        """Return the k most similar datasets with respect to self.metric

        Parameters
        ----------
        x : pandas.Series
            A pandas Series object with the metafeatures for one dataset

        k : int
            Number of k nearest datasets which are returned. If k == -1,
            return all dataset sorted by similarity.

        return_distance : bool, optional. Defaults to False
            If true, distances to the new dataset will be returned.

        Returns
        -------
        list
            Names of the most similar datasets, sorted by similarity

        list
            Sorted distances. Only returned if return_distances is set to True.
        """
        assert type(x) == pd.Series
        if k < -1 or k == 0:
            raise ValueError("Number of neighbors k cannot be zero or negative.")
        elif k == -1:
            k = self.num_datasets

        X_train = self.scaler.transform(self.metafeatures)
        x = x.values.reshape((1, -1))
        x = self.scaler.transform(x)
        self._nearest_neighbors.fit(X_train)
        distances, neighbor_indices = self._nearest_neighbors.kneighbors(
            x, n_neighbors=k, return_distance=True
        )

        assert k == neighbor_indices.shape[1]

        return_value = [
            self.metafeatures.index[i]
            # Neighbor indices is 2d, each row is the indices for one
            # dataset in x.
            for i in neighbor_indices[0]
        ]

        if return_distance is False:
            return return_value
        else:
            return return_value, distances[0]

    def kBestSuggestions(self, x, k=1, exclude_double_configurations=True):
        assert type(x) == pd.Series
        if k < -1 or k == 0:
            raise ValueError("Number of neighbors k cannot be zero or negative.")
        nearest_datasets, distances = self.kNearestDatasets(x, -1, return_distance=True)

        kbest = []

        added_configurations = set()
        for dataset_name, distance in zip(nearest_datasets, distances):
            best_configuration = self.best_configuration_per_dataset[dataset_name]

            if best_configuration is None:
                self.logger.info(
                    "Found no best configuration for instance %s" % dataset_name
                )
                continue

            if exclude_double_configurations:
                if best_configuration not in added_configurations:
                    added_configurations.add(best_configuration)
                    kbest.append((dataset_name, distance, best_configuration))
            else:
                kbest.append((dataset_name, distance, best_configuration))

            if k != -1 and len(kbest) >= k:
                break

        if k == -1:
            k = len(kbest)
        return kbest[:k]



import sklearn.gaussian_process

from autosklearn.pipeline.components.regression.gaussian_process import GaussianProcess

from .test_base import BaseRegressionComponentTest


class GaussianProcessComponentTest(BaseRegressionComponentTest):

    __test__ = True

    res = dict()
    res["default_boston_le_ge"] = [0.6, 0.2]
    res["default_boston_places"] = 1
    res["default_boston_iterative"] = None
    res["default_boston_sparse"] = None
    res["default_boston_iterative_sparse"] = None
    res["default_diabetes"] = -0.017256687184589836
    res["default_diabetes_iterative"] = None
    res["default_diabetes_sparse"] = None
    res["default_diabetes_iterative_sparse"] = None

    sk_mod = sklearn.gaussian_process.GaussianProcessRegressor

    module = GaussianProcess

    """
    # Leave this here for future reference
    # My machine: 0.574913739659292
    # travis-ci: 0.49562471963524557
    self.assertLessEqual(
        sklearn.metrics.r2_score(y_true=targets, y_pred=predictions),
        0.6)
    self.assertGreaterEqual(
        sklearn.metrics.r2_score(y_true=targets, y_pred=predictions),
        0.4)
    """



from typing import Dict, Optional

import numpy as np
import sklearn.metrics

from autosklearn.pipeline.constants import SPARSE
from autosklearn.pipeline.util import (
    _test_classifier,
    _test_classifier_iterative_fit,
    _test_classifier_predict_proba,
)

import unittest

from test.test_pipeline.ignored_warnings import classifier_warnings, ignore_warnings


class BaseClassificationComponentTest(unittest.TestCase):
    # Magic command to not run tests on base class
    __test__ = False

    res = None

    module = None
    sk_module = None
    # Hyperparameter which is increased by iterative_fit
    step_hyperparameter = None

    def test_default_iris(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        for i in range(2):
            predictions, targets, n_calls = _test_classifier(
                dataset="iris", classifier=self.module
            )
            self.assertAlmostEqual(
                self.res["default_iris"],
                sklearn.metrics.accuracy_score(targets, predictions),
                places=self.res.get("default_iris_places", 7),
            )

            if self.res.get("iris_n_calls"):
                self.assertEqual(self.res["iris_n_calls"], n_calls)

    def test_get_max_iter(self):
        if self.__class__ == BaseClassificationComponentTest:
            return

        if not hasattr(self.module, "iterative_fit"):
            return

        self.module.get_max_iter()

    def test_default_iris_iterative_fit(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        if not hasattr(self.module, "iterative_fit"):
            return

        for i in range(2):
            predictions, targets, classifier = _test_classifier_iterative_fit(
                dataset="iris", classifier=self.module
            )
            self.assertAlmostEqual(
                self.res["default_iris_iterative"],
                sklearn.metrics.accuracy_score(targets, predictions),
                places=self.res.get("default_iris_iterative_places", 7),
            )

            if self.step_hyperparameter is not None:
                self.assertEqual(
                    getattr(classifier.estimator, self.step_hyperparameter["name"]),
                    self.res.get(
                        "iris_iterative_n_iter", self.step_hyperparameter["value"]
                    ),
                )

    def test_default_iris_predict_proba(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        for _ in range(2):
            predictions, targets = _test_classifier_predict_proba(
                dataset="iris", classifier=self.module
            )
            self.assertAlmostEqual(
                self.res["default_iris_proba"],
                sklearn.metrics.log_loss(targets, predictions),
                places=self.res.get("default_iris_proba_places", 7),
            )

    def test_default_iris_sparse(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        if SPARSE not in self.module.get_properties()["input"]:
            return

        for i in range(2):
            predictions, targets, _ = _test_classifier(
                dataset="iris", classifier=self.module, sparse=True
            )
            self.assertAlmostEqual(
                self.res["default_iris_sparse"],
                sklearn.metrics.accuracy_score(targets, predictions),
                places=self.res.get("default_iris_sparse_places", 7),
            )

    def test_default_digits_binary(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        for i in range(2):
            predictions, targets, _ = _test_classifier(
                classifier=self.module, dataset="digits", sparse=False, make_binary=True
            )
            self.assertAlmostEqual(
                self.res["default_digits_binary"],
                sklearn.metrics.accuracy_score(targets, predictions),
                places=self.res.get("default_digits_binary_places", 7),
            )

    def test_default_digits(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        for i in range(2):
            predictions, targets, n_calls = _test_classifier(
                dataset="digits", classifier=self.module
            )
            self.assertAlmostEqual(
                self.res["default_digits"],
                sklearn.metrics.accuracy_score(targets, predictions),
                places=self.res.get("default_digits_places", 7),
            )

            if self.res.get("digits_n_calls"):
                self.assertEqual(self.res["digits_n_calls"], n_calls)

    def test_default_digits_iterative_fit(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        if not hasattr(self.module, "iterative_fit"):
            return

        for i in range(2):
            predictions, targets, classifier = _test_classifier_iterative_fit(
                dataset="digits", classifier=self.module
            )
            self.assertAlmostEqual(
                self.res["default_digits_iterative"],
                sklearn.metrics.accuracy_score(targets, predictions),
                places=self.res.get("default_digits_iterative_places", 7),
            )

            if self.step_hyperparameter is not None:
                self.assertEqual(
                    getattr(classifier.estimator, self.step_hyperparameter["name"]),
                    self.res.get(
                        "digits_iterative_n_iter", self.step_hyperparameter["value"]
                    ),
                )

    def test_default_digits_multilabel(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        if not self.module.get_properties()["handles_multilabel"]:
            return

        for _ in range(2):
            predictions, targets, _ = _test_classifier(
                classifier=self.module, dataset="digits", make_multilabel=True
            )

            score = sklearn.metrics.precision_score(
                targets, predictions, average="macro", zero_division=0
            )
            self.assertAlmostEqual(
                self.res["default_digits_multilabel"],
                score,
                places=self.res.get("default_digits_multilabel_places", 7),
            )

    def test_default_digits_multilabel_predict_proba(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        if not self.module.get_properties()["handles_multilabel"]:
            return

        for i in range(2):
            predictions, targets = _test_classifier_predict_proba(
                classifier=self.module, make_multilabel=True
            )
            self.assertEqual(predictions.shape, ((50, 3)))
            self.assertAlmostEqual(
                self.res["default_digits_multilabel_proba"],
                sklearn.metrics.roc_auc_score(targets, predictions, average="macro"),
                places=self.res.get("default_digits_multilabel_proba_places", 7),
            )

    def test_target_algorithm_multioutput_multiclass_support(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        if not self.module.get_properties()["handles_multiclass"]:
            return
        elif self.sk_module is not None:
            cls = self.sk_module
            X = np.random.random((10, 10))
            y = np.random.randint(0, 1, size=(10, 10))
            self.assertRaisesRegex(
                ValueError, "bad input shape \\(10, 10\\)", cls.fit, X, y
            )
        else:
            return

    def test_module_idempotent(self):
        """Fitting twice with the same config gives the same model params.

        This is only valid when the random_state passed is an int. If a
        RandomState object is passed then repeated calls to fit will have
        different results. See the section on "Controlling Randomness" in the
        sklearn docs.

        https://scikit-learn.org/0.24/common_pitfalls.html#controlling-randomness
        """
        if self.__class__ == BaseClassificationComponentTest:
            return

        classifier_cls = self.module

        X = np.array(
            [
                [0, 0],
                [0, 1],
                [1, 0],
                [1, 1],
                [0, 0],
                [0, 1],
                [1, 0],
                [1, 1],
                [0, 0],
                [0, 1],
                [1, 0],
                [1, 1],
                [0, 0],
                [0, 1],
                [1, 0],
                [1, 1],
            ]
        )
        y = np.array(
            [
                0,
                1,
                1,
                0,
                0,
                1,
                1,
                0,
                0,
                1,
                1,
                0,
                0,
                1,
                1,
                0,
            ]
        )

        # There are certain errors we ignore so we wrap this in a function
        def fitted_params(model) -> Optional[Dict]:
            """
            Returns the params if fitted successfully, else None if an
            acceptable error occurs
            """
            # We are okay with Numerical in Quadractic disciminant analysis
            def is_QDA_error(err):
                return "Numerical problems in QDA" in err.args[0]

            # We are okay if the BaseClassifier in AdaBoostClassifier is worse
            # than random so no ensemble can be fit
            def is_AdaBoostClassifier_error(err):
                return (
                    "BaseClassifier in AdaBoostClassifier ensemble is worse"
                    + " than random, ensemble can not be fit."
                    in err.args[0]
                )

            def is_unset_param_raw_predictions_val_error(err):
                return (
                    "local variable 'raw_predictions_val' referenced before"
                    + " assignment"
                    in err.args[0]
                )

            try:
                with ignore_warnings(classifier_warnings):
                    model.fit(X.copy(), y.copy())
            except ValueError as e:
                if is_AdaBoostClassifier_error(e) or is_QDA_error(e):
                    return None
            except UnboundLocalError as e:
                if is_unset_param_raw_predictions_val_error(e):
                    return None

            return model.estimator.get_params()

        # We ignore certain keys when comparing
        param_keys_ignored = ["base_estimator"]

        # We use the default config + sampled ones
        configuration_space = classifier_cls.get_hyperparameter_search_space()

        default = configuration_space.get_default_configuration()
        sampled = [configuration_space.sample_configuration() for _ in range(2)]

        for seed, config in enumerate([default] + sampled):
            model_args = {"random_state": seed, **config}
            classifier = classifier_cls(**model_args)

            # Get the parameters on the first and second fit with config params
            params_first = fitted_params(classifier)
            if hasattr(classifier.estimator, "random_state"):
                rs_1 = classifier.random_state
                rs_estimator_1 = classifier.estimator.random_state

            params_second = fitted_params(classifier)
            if hasattr(classifier.estimator, "random_state"):
                rs_2 = classifier.random_state
                rs_estimator_2 = classifier.estimator.random_state

            # An acceptable error occured, skip to next sample
            if params_first is None or params_second is None:
                continue

            # Remove keys we don't wish to include in the comparison
            for params in [params_first, params_second]:
                for key in param_keys_ignored:
                    if key in params:
                        del params[key]

            # They should have equal parameters
            self.assertEqual(
                params_first, params_second, f"Failed with model args {model_args}"
            )
            if hasattr(classifier.estimator, "random_state"):
                assert all(
                    [
                        seed == random_state
                        for random_state in [rs_1, rs_estimator_1, rs_2, rs_estimator_2]
                    ]
                )



import sklearn.utils


def get_random_metric(random_state=1):
    random_state = sklearn.utils.check_random_state(random_state)

    def _random(d1, d2):
        return random_state.random_sample()

    return _random



from abc import ABCMeta, abstractmethod
from typing import Dict, List, Tuple, Union

import numpy as np

from sklearn.pipeline import Pipeline


class AbstractEnsemble(object, metaclass=ABCMeta):
    @abstractmethod
    def fit(
        self,
        base_models_predictions: np.ndarray,
        true_targets: np.ndarray,
        model_identifiers: List[Tuple[int, int, float]],
    ) -> "AbstractEnsemble":
        """Fit an ensemble given predictions of base models and targets.

        Ensemble building maximizes performance (in contrast to
        hyperparameter optimization)!

        Parameters
        ----------
        base_models_predictions : array of shape = [n_base_models, n_data_points, n_targets]
            n_targets is the number of classes in case of classification,
            n_targets is 0 or 1 in case of regression

        true_targets : array of shape [n_targets]

        model_identifiers : identifier for each base model.
            Can be used for practical text output of the ensemble.

        Returns
        -------
        self

        """
        pass

    @abstractmethod
    def predict(self, base_models_predictions: Union[np.ndarray, List[np.ndarray]]) -> np.ndarray:
        """Create ensemble predictions from the base model predictions.

        Parameters
        ----------
        base_models_predictions : array of shape = [n_base_models, n_data_points, n_targets]
            Same as in the fit method.

        Returns
        -------
        array : [n_data_points]
        """
        self

    @abstractmethod
    def get_models_with_weights(self, models: Dict[str, Pipeline]) -> List[Tuple[float, Pipeline]]:
        """Return a list of (weight, model) pairs

        Parameters
        ----------
        models : dict {identifier : model object}
            The identifiers are the same as the one presented to the fit()
            method. Models can be used for nice printing.

        Returns
        -------
        array : [(weight_1, model_1), ..., (weight_n, model_n)]
        """

    @abstractmethod
    def get_selected_model_identifiers(self) -> List[Tuple[int, int, float]]:
        """Return identifiers of models in the ensemble.

        This includes models which have a weight of zero!

        Returns
        -------
        list
        """

    @abstractmethod
    def get_validation_performance(self) -> float:
        """Return validation performance of ensemble.

        Return
        ------
        float
        """



from typing import Optional

from autosklearn.util.logging_ import PicklableClientLogger

from unittest.mock import Mock

MOCKNAME = "mock"
MOCKHOST = "mockhost"
MOCKPORT = 9020


class MockLogger(PicklableClientLogger):
    """Should not be used for testing the actual loggers functionality

    Overwrites all methods with mock objects that can be queries
    * All logging methods do nothing
    * isEnabledFor returns True for everything as it's part of the logging config we
      don't have access to
    * __setstate__ and __getstate__ remain the same and are not mocked
    """

    def __init__(
        self,
        name: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        self.name = name or MOCKNAME
        self.host = host or MOCKHOST
        self.port = port or MOCKPORT

        # Overwrite the logging implementations with mocks
        self.debug = Mock(return_value=None)  # type: ignore
        self.info = Mock(return_value=None)  # type: ignore
        self.warning = Mock(return_value=None)  # type: ignore
        self.error = Mock(return_value=None)  # type: ignore
        self.exception = Mock(return_value=None)  # type: ignore
        self.critical = Mock(return_value=None)  # type: ignore
        self.log = Mock(return_value=None)  # type: ignore
        self.isEnabledFor = Mock(return_value=True)  # type: ignore



# -*- encoding: utf-8 -*-
import os
import sys

import pkg_resources

from autosklearn.__version__ import __version__  # noqa (imported but unused)
from autosklearn.util import dependencies

requirements = pkg_resources.resource_string("autosklearn", "requirements.txt")
requirements = requirements.decode("utf-8")

dependencies.verify_packages(requirements)

if os.name != "posix":
    raise ValueError(
        "Detected unsupported operating system: %s. Please check "
        "the compability information of auto-sklearn: https://automl.github.io"
        "/auto-sklearn/stable/installation.html#windows-osx-compability" % sys.platform
    )

if sys.version_info < (3, 6):
    raise ValueError(
        "Unsupported python version %s found. Auto-sklearn requires Python "
        "3.6 or higher." % sys.version_info
    )



from typing import Optional

import warnings

from ConfigSpace.conditions import EqualsCondition
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformIntegerHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, UNSIGNED_DATA
from autosklearn.util.common import check_for_bool, check_none


class FastICA(AutoSklearnPreprocessingAlgorithm):
    def __init__(self, algorithm, whiten, fun, n_components=None, random_state=None):
        self.algorithm = algorithm
        self.whiten = whiten
        self.fun = fun
        self.n_components = n_components

        self.random_state = random_state

    def fit(self, X, Y=None):
        import sklearn.decomposition

        self.whiten = check_for_bool(self.whiten)
        if check_none(self.n_components):
            self.n_components = None
        else:
            self.n_components = int(self.n_components)

        self.preprocessor = sklearn.decomposition.FastICA(
            n_components=self.n_components,
            algorithm=self.algorithm,
            fun=self.fun,
            whiten=self.whiten,
            random_state=self.random_state,
        )
        # Make the RuntimeWarning an Exception!
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "error", message="array must not contain infs or NaNs"
            )
            try:
                self.preprocessor.fit(X)
            except ValueError as e:
                if "array must not contain infs or NaNs" in e.args[0]:
                    raise ValueError(
                        "Bug in scikit-learn: "
                        "https://github.com/scikit-learn/scikit-learn/pull/2738"
                    )

        return self

    def transform(self, X):
        if self.preprocessor is None:
            raise NotImplementedError()
        return self.preprocessor.transform(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "FastICA",
            "name": "Fast Independent Component Analysis",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "is_deterministic": False,
            "input": (DENSE, UNSIGNED_DATA),
            "output": (INPUT, UNSIGNED_DATA),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        cs = ConfigurationSpace()

        n_components = UniformIntegerHyperparameter(
            "n_components", 10, 2000, default_value=100
        )
        algorithm = CategoricalHyperparameter(
            "algorithm", ["parallel", "deflation"], "parallel"
        )
        whiten = CategoricalHyperparameter("whiten", ["False", "True"], "False")
        fun = CategoricalHyperparameter("fun", ["logcosh", "exp", "cube"], "logcosh")
        cs.add_hyperparameters([n_components, algorithm, whiten, fun])

        cs.add_condition(EqualsCondition(n_components, whiten, "True"))

        return cs



from typing import Dict, Optional

import os
from collections import OrderedDict

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import CategoricalHyperparameter
from sklearn.base import BaseEstimator

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.data_preprocessing.rescaling.abstract_rescaling import (  # noqa: E501
    Rescaling,
)

from ...base import (
    AutoSklearnChoice,
    AutoSklearnPreprocessingAlgorithm,
    ThirdPartyComponents,
    _addons,
    find_components,
)

rescaling_directory = os.path.split(__file__)[0]
_rescalers = find_components(
    __package__, rescaling_directory, AutoSklearnPreprocessingAlgorithm
)
additional_components = ThirdPartyComponents(AutoSklearnPreprocessingAlgorithm)
_addons["data_preprocessing.rescaling"] = additional_components


def add_rescaler(rescaler: Rescaling) -> None:
    additional_components.add_component(rescaler)


class RescalingChoice(AutoSklearnChoice):
    @classmethod
    def get_components(cls: BaseEstimator) -> Dict[str, BaseEstimator]:
        components: Dict[str, BaseEstimator] = OrderedDict()
        components.update(_rescalers)
        components.update(additional_components.components)
        return components

    def get_hyperparameter_search_space(
        self,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
        default: Optional[str] = None,
        include: Optional[Dict[str, str]] = None,
        exclude: Optional[Dict[str, str]] = None,
    ) -> ConfigurationSpace:
        cs = ConfigurationSpace()

        if dataset_properties is None:
            dataset_properties = {}

        # Compile a list of legal preprocessors for this problem
        available_preprocessors = self.get_available_components(
            dataset_properties=dataset_properties, include=include, exclude=exclude
        )

        if len(available_preprocessors) == 0:
            raise ValueError("No rescalers found, please add any rescaling component.")

        if default is None:
            defaults = ["standardize", "none", "minmax", "normalize"]
            for default_ in defaults:
                if default_ in available_preprocessors:
                    default = default_
                    break

        preprocessor = CategoricalHyperparameter(
            "__choice__", list(available_preprocessors.keys()), default_value=default
        )
        cs.add_hyperparameter(preprocessor)
        for name in available_preprocessors:
            preprocessor_configuration_space = available_preprocessors[
                name
            ].get_hyperparameter_search_space(
                feat_type=feat_type, dataset_properties=dataset_properties
            )
            parent_hyperparameter = {"parent": preprocessor, "value": name}
            cs.add_configuration_space(
                name,
                preprocessor_configuration_space,
                parent_hyperparameter=parent_hyperparameter,
            )

        self.configuration_space = cs
        self.dataset_properties = dataset_properties
        return cs

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        return self.choice.transform(X)



# -*- encoding: utf-8 -*-
from typing import Dict, Optional, Union, cast

import numpy as np
import pandas as pd
from scipy import sparse

from autosklearn.constants import (
    BINARY_CLASSIFICATION,
    MULTICLASS_CLASSIFICATION,
    MULTILABEL_CLASSIFICATION,
    MULTIOUTPUT_REGRESSION,
    REGRESSION,
)
from autosklearn.data.abstract_data_manager import AbstractDataManager
from autosklearn.data.validation import SUPPORTED_FEAT_TYPES, SUPPORTED_TARGET_TYPES


class XYDataManager(AbstractDataManager):
    def __init__(
        self,
        X: SUPPORTED_FEAT_TYPES,
        y: SUPPORTED_TARGET_TYPES,
        X_test: Optional[SUPPORTED_FEAT_TYPES],
        y_test: Optional[SUPPORTED_TARGET_TYPES],
        task: int,
        feat_type: Dict[Union[str, int], str],
        dataset_name: str,
    ):
        super(XYDataManager, self).__init__(dataset_name)

        self.info["task"] = task
        if sparse.issparse(X):
            self.info["is_sparse"] = 1
            self.info["has_missing"] = np.all(
                np.isfinite(cast(sparse.csr_matrix, X).data)
            )
        else:
            self.info["is_sparse"] = 0
            if hasattr(X, "iloc"):
                self.info["has_missing"] = cast(pd.DataFrame, X).isnull().values.any()
            else:
                self.info["has_missing"] = np.all(np.isfinite(X))

        label_num = {
            REGRESSION: 1,
            BINARY_CLASSIFICATION: 2,
            MULTIOUTPUT_REGRESSION: np.shape(y)[-1],
            MULTICLASS_CLASSIFICATION: len(np.unique(y)),
            MULTILABEL_CLASSIFICATION: np.shape(y)[-1],
        }

        self.info["label_num"] = label_num[task]

        self.data["X_train"] = X
        self.data["Y_train"] = y
        if X_test is not None:
            self.data["X_test"] = X_test
        if y_test is not None:
            self.data["Y_test"] = y_test

        if isinstance(feat_type, dict):
            self.feat_type = feat_type
        else:
            raise ValueError(
                "Unsupported feat_type provided. We expect the user to "
                "provide a Dict[str, str] mapping from column to categorical/ "
                "numerical."
            )

        # TODO: try to guess task type!

        if len(np.shape(y)) > 2:
            raise ValueError(
                "y must not have more than two dimensions, "
                "but has %d." % len(np.shape(y))
            )

        if np.shape(X)[0] != np.shape(y)[0]:
            raise ValueError(
                "X and y must have the same number of "
                "datapoints, but have %d and %d." % (np.shape(X)[0], np.shape(y)[0])
            )



"""Test things related to how AutoML compresses the dataset size"""
from typing import Any, Callable, Dict

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix

from autosklearn.automl import AutoML
from autosklearn.constants import BINARY_CLASSIFICATION

from pytest_cases import parametrize
from unittest.mock import patch

from test.util import skip


@parametrize("dataset_compression", [{"methods": ["precision", "subsample"]}])
def test_fit_performs_dataset_compression_without_precision_when_int(
    dataset_compression: Dict,
    make_automl: Callable[..., AutoML],
) -> None:
    """
    Parameters
    ----------
    dataset_compression: Dict
        The dataset_compression arg with "precision" set in it

    Fixtures
    --------
    make_automl: Callable[..., AutoML]
        Makes an automl instance


    Expects
    -------
    * Should call reduce_dataset_size_if_too_large
    * "precision" should have been removed from the "methods" passed to the keyword
        argument "operations" of `reduce_dataset_size_if_too_large`.

    Note
    ----
    * Only done with int's as we can't reduce precision of ints in a meaningful way
    """
    X = np.ones((100, 10), dtype=int)
    y = np.random.random((100,))

    auto = make_automl(dataset_compression=dataset_compression)

    with patch(
        "autosklearn.automl.reduce_dataset_size_if_too_large", return_value=(X, y)
    ) as mck:
        # To prevent fitting anything we use `only_return_configuration_space`
        auto.fit(X, y, only_return_configuration_space=True, task=BINARY_CLASSIFICATION)

    assert mck.call_count == 1

    args, kwargs = mck.call_args
    assert kwargs["operations"] == ["subsample"]


@parametrize(
    "X_type",
    [
        np.ndarray,
        csr_matrix,
        skip(
            list,
            "dataset_compression does not support pandas types yet and list gets"
            " converted in InputValidator",
        ),
        skip(pd.DataFrame, "dataset_compression does not support pandas types yet"),
    ],
)
@parametrize(
    "y_type",
    [
        np.ndarray,
        skip(csr_matrix, "See TODO note in `test_fit_performs_dataset_compression`"),
        list,
        skip(pd.DataFrame, "dataset_compression does not support pandas types yet"),
        skip(pd.Series, "dataset_compression does not support pandas types yet"),
    ],
)
def test_fit_performs_dataset_compression(
    X_type: Any,
    y_type: Any,
    make_automl: Callable[..., AutoML],
    make_data: Callable[..., Any],
) -> None:
    """
    Parameters
    ----------
    mock_reduce_dataset: MagicMock
        A mock function to view call

    X_type: Union[np.ndarray, csr_matrix, list, pd.Dataframe]
        Feature to reduce

    y_type: Union[np.ndarray, csr_matrix, list, pd.Series, pd.Dataframe]
        Target to reduce (regression values)

    Fixtures
    --------
    make_automl: Callable[..., AutoML]
        Factory to make automl instance

    make_data: Callable
        Factory to make data

    Expects
    -------
    * Should call reduce_dataset_size_if_too_large

    # TODO not sure how to keep function behaviour and just use the mock object so
    # that we can assert it was called.
    #
    # * `fit` will convert sparse `y`
    # * This gets passed to `reduce_dataset_size_if_too_large`
    # * The de-sparsified `y` is required for the datamanager later on
    #
    # Mocking away the functionality and just returning the X, y we see here will means
    # that the datamanager will get the sparse y and crash, hence we manually convert
    # here
    """
    X, y = make_data(types=(X_type, y_type))

    auto = make_automl(dataset_compression=True)

    with patch(
        "autosklearn.automl.reduce_dataset_size_if_too_large", return_value=(X, y)
    ) as mck:
        # To prevent fitting anything we use `only_return_configuration_space`
        auto.fit(X, y, only_return_configuration_space=True, task=BINARY_CLASSIFICATION)

    assert mck.called
    del auto



from __future__ import annotations

from typing import Any, Dict, List, Mapping, Optional, Union

import hashlib
import json
import os
import pathlib
import pickle

import dask.distributed
import numpy as np
import pandas as pd
import scipy.sparse
import sklearn
from ConfigSpace import Configuration

import autosklearn
import autosklearn.experimental.selector
from autosklearn.classification import AutoSklearnClassifier
from autosklearn.ensembles.abstract_ensemble import AbstractEnsemble
from autosklearn.ensembles.ensemble_selection import EnsembleSelection
from autosklearn.metrics import Scorer, accuracy, balanced_accuracy, log_loss, roc_auc

selector_metrics = (balanced_accuracy, roc_auc, log_loss)


class SmacObjectCallback:
    def __init__(self, portfolio):
        self.portfolio = portfolio

    def __call__(
        self,
        scenario_dict,
        seed,
        ta,
        ta_kwargs,
        metalearning_configurations,
        n_jobs,
        dask_client,
        multi_objective_algorithm,
        multi_objective_kwargs,
    ):
        from smac.facade.smac_ac_facade import SMAC4AC
        from smac.intensification.simple_intensifier import SimpleIntensifier
        from smac.runhistory.runhistory2epm import RunHistory2EPM4LogCost
        from smac.scenario.scenario import Scenario

        scenario = Scenario(scenario_dict)

        initial_configurations = []
        for member in self.portfolio.values():
            try:
                hp_names = scenario.cs.get_hyperparameter_names()
                _member = {key: member[key] for key in member if key in hp_names}
                initial_configurations.append(
                    Configuration(configuration_space=scenario.cs, values=_member)
                )
            except ValueError:
                pass

        rh2EPM = RunHistory2EPM4LogCost
        return SMAC4AC(
            scenario=scenario,
            rng=seed,
            runhistory2epm=rh2EPM,
            tae_runner=ta,
            tae_runner_kwargs=ta_kwargs,
            initial_configurations=initial_configurations,
            intensifier=SimpleIntensifier,
            run_id=seed,
            n_jobs=n_jobs,
            dask_client=dask_client,
            multi_objective_algorithm=multi_objective_algorithm,
            multi_objective_kwargs=multi_objective_kwargs,
        )


class SHObjectCallback:
    def __init__(self, budget_type, eta, initial_budget, portfolio):
        self.budget_type = budget_type
        self.eta = eta
        self.initial_budget = initial_budget
        self.portfolio = portfolio

    def __call__(
        self,
        scenario_dict,
        seed,
        ta,
        ta_kwargs,
        metalearning_configurations,
        n_jobs,
        dask_client,
        multi_objective_algorithm,
        multi_objective_kwargs,
    ):
        from smac.facade.smac_ac_facade import SMAC4AC
        from smac.intensification.successive_halving import SuccessiveHalving
        from smac.runhistory.runhistory2epm import RunHistory2EPM4LogCost
        from smac.scenario.scenario import Scenario

        scenario = Scenario(scenario_dict)

        initial_configurations = []
        for member in self.portfolio.values():
            try:
                hp_names = scenario.cs.get_hyperparameter_names()
                _member = {key: member[key] for key in member if key in hp_names}
                initial_configurations.append(
                    Configuration(configuration_space=scenario.cs, values=_member)
                )
            except ValueError:
                pass

        rh2EPM = RunHistory2EPM4LogCost
        ta_kwargs["budget_type"] = self.budget_type

        smac4ac = SMAC4AC(
            scenario=scenario,
            rng=seed,
            runhistory2epm=rh2EPM,
            tae_runner=ta,
            tae_runner_kwargs=ta_kwargs,
            initial_configurations=initial_configurations,
            run_id=seed,
            intensifier=SuccessiveHalving,
            intensifier_kwargs={
                "initial_budget": self.initial_budget,
                "max_budget": 100,
                "eta": self.eta,
                "min_chall": 1,
            },
            dask_client=dask_client,
            n_jobs=n_jobs,
            multi_objective_algorithm=multi_objective_algorithm,
            multi_objective_kwargs=multi_objective_kwargs,
        )
        smac4ac.solver.epm_chooser.min_samples_model = int(
            len(scenario.cs.get_hyperparameters()) / 2
        )
        return smac4ac


class AutoSklearn2Classifier(AutoSklearnClassifier):
    def __init__(
        self,
        time_left_for_this_task: int = 3600,
        per_run_time_limit=None,
        ensemble_size: int | None = None,
        ensemble_class: AbstractEnsemble | None = EnsembleSelection,
        ensemble_kwargs: Dict[str, Any] | None = None,
        ensemble_nbest: Union[float, int] = 50,
        max_models_on_disc: int = 50,
        seed: int = 1,
        memory_limit: int = 3072,
        tmp_folder: Optional[str] = None,
        delete_tmp_folder_after_terminate: bool = True,
        n_jobs: Optional[int] = None,
        dask_client: Optional[dask.distributed.Client] = None,
        disable_evaluator_output: bool = False,
        smac_scenario_args: Optional[Dict[str, Any]] = None,
        logging_config: Optional[Dict[str, Any]] = None,
        metric: Optional[Scorer] = None,
        scoring_functions: Optional[List[Scorer]] = None,
        load_models: bool = True,
        dataset_compression: Union[bool, Mapping[str, Any]] = True,
        allow_string_features: bool = True,
    ):

        """
        Parameters
        ----------
        time_left_for_this_task : int, optional (default=3600)
            Time limit in seconds for the search of appropriate
            models. By increasing this value, *auto-sklearn* has a higher
            chance of finding better models.

        per_run_time_limit : int, optional (default=1/10 of time_left_for_this_task)
            Time limit for a single call to the machine learning model.
            Model fitting will be terminated if the machine learning
            algorithm runs over the time limit. Set this value high enough so
            that typical machine learning algorithms can be fit on the
            training data.

        ensemble_size : int, optional
            Number of models added to the ensemble built by *Ensemble
            selection from libraries of models*. Models are drawn with
            replacement. If set to ``0`` no ensemble is fit.

            Deprecated - will be removed in Auto-sklearn 0.16. Please pass
            this argument via ``ensemble_kwargs={"ensemble_size": int}``
            if you want to change the ensemble size for ensemble selection.

        ensemble_class : Type[AbstractEnsemble], optional (default=EnsembleSelection)
            Class implementing the post-hoc ensemble algorithm. Set to
            ``None`` to disable ensemble building or use ``SingleBest``
            to obtain only use the single best model instead of an
            ensemble.

        ensemble_kwargs : Dict, optional
            Keyword arguments that are passed to the ensemble class upon
            initialization.

        max_models_on_disc: int, optional (default=50),
            Defines the maximum number of models that are kept in the disc.
            The additional number of models are permanently deleted. Due to the
            nature of this variable, it sets the upper limit on how many models
            can be used for an ensemble.
            It must be an integer greater or equal than 1.
            If set to None, all models are kept on the disc.

        seed : int, optional (default=1)
            Used to seed SMAC. Will determine the output file names.

        memory_limit : int, optional (3072)
            Memory limit in MB for the machine learning algorithm.
            `auto-sklearn` will stop fitting the machine learning algorithm if
            it tries to allocate more than ``memory_limit`` MB.

            **Important notes:**

            * If ``None`` is provided, no memory limit is set.
            * In case of multi-processing, ``memory_limit`` will be *per job*, so the total usage is
              ``n_jobs x memory_limit``.
            * The memory limit also applies to the ensemble creation process.

        tmp_folder : string, optional (None)
            folder to store configuration output and log files, if ``None``
            automatically use ``/tmp/autosklearn_tmp_$pid_$random_number``

        delete_tmp_folder_after_terminate: string, optional (True)
            remove tmp_folder, when finished. If tmp_folder is None
            tmp_dir will always be deleted

        n_jobs : int, optional, experimental
            The number of jobs to run in parallel for ``fit()``. ``-1`` means
            using all processors.

            **Important notes**:

            * By default, Auto-sklearn uses one core.
            * Ensemble building is not affected by ``n_jobs`` but can be controlled by the number
              of models in the ensemble.
            * ``predict()`` is not affected by ``n_jobs`` (in contrast to most scikit-learn models)
            * If ``dask_client`` is ``None``, a new dask client is created.

        dask_client : dask.distributed.Client, optional
            User-created dask client, can be used to start a dask cluster and then
            attach auto-sklearn to it.

        disable_evaluator_output: bool or list, optional (False)
            If True, disable model and prediction output. Cannot be used
            together with ensemble building. ``predict()`` cannot be used when
            setting this True. Can also be used as a list to pass more
            fine-grained information on what to save. Allowed elements in the
            list are:

            * ``'y_optimization'`` : do not save the predictions for the
              optimization/validation set, which would later on be used to build
              an ensemble.
            * ``model`` : do not save any model files

        smac_scenario_args : dict, optional (None)
            Additional arguments inserted into the scenario of SMAC. See the
            `SMAC documentation <https://automl.github.io/SMAC3/main/api/smac.scenario.scenario.html#module-smac.scenario.scenario>`_
            for a list of available arguments.

        logging_config : dict, optional (None)
            dictionary object specifying the logger configuration. If None,
            the default logging.yaml file is used, which can be found in
            the directory ``util/logging.yaml`` relative to the installation.

        metric : Scorer, optional (None)
            An instance of :class:`autosklearn.metrics.Scorer` as created by
            :meth:`autosklearn.metrics.make_scorer`. These are the `Built-in
            Metrics`_.
            If None is provided, a default metric is selected depending on the task.

        scoring_functions : List[Scorer], optional (None)
            List of scorers which will be calculated for each pipeline and results will be
            available via ``cv_results``

        load_models : bool, optional (True)
            Whether to load the models after fitting Auto-sklearn.

        Attributes
        ----------

        cv_results_ : dict of numpy (masked) ndarrays
            A dict with keys as column headers and values as columns, that can be
            imported into a pandas ``DataFrame``.

            Not all keys returned by scikit-learn are supported yet.

        """  # noqa (links are too long)
        self.required_training = False  # Boolean to indicate if selectors were trained.
        include_estimators = [
            "extra_trees",
            "passive_aggressive",
            "random_forest",
            "sgd",
            "gradient_boosting",
            "mlp",
        ]
        include_preprocessors = ["no_preprocessing"]
        include = {
            "classifier": include_estimators,
            "feature_preprocessor": include_preprocessors,
        }
        self.train_selectors(selected_metric=metric)
        super().__init__(
            time_left_for_this_task=time_left_for_this_task,
            per_run_time_limit=per_run_time_limit,
            initial_configurations_via_metalearning=0,
            ensemble_size=ensemble_size,
            ensemble_class=ensemble_class,
            ensemble_kwargs=ensemble_kwargs,
            ensemble_nbest=ensemble_nbest,
            max_models_on_disc=max_models_on_disc,
            seed=seed,
            memory_limit=memory_limit,
            include=include,
            exclude=None,
            resampling_strategy=None,
            resampling_strategy_arguments=None,
            tmp_folder=tmp_folder,
            delete_tmp_folder_after_terminate=delete_tmp_folder_after_terminate,
            n_jobs=n_jobs,
            dask_client=dask_client,
            disable_evaluator_output=disable_evaluator_output,
            get_smac_object_callback=None,
            smac_scenario_args=smac_scenario_args,
            logging_config=logging_config,
            metadata_directory=None,
            metric=metric,
            scoring_functions=scoring_functions,
            load_models=load_models,
            allow_string_features=allow_string_features,
        )

    def train_selectors(self, selected_metric=None):
        self.selector_metrics = (balanced_accuracy, roc_auc, log_loss)
        self.selector_files = {}
        self.this_directory = pathlib.Path(__file__).resolve().parent

        if selected_metric is not None:
            metric_list = [selected_metric]
        else:
            metric_list = self.selector_metrics

        for metric in metric_list:
            training_data_file = (
                self.this_directory / metric.name / "askl2_training_data.json"
            )
            with open(training_data_file) as fh:
                training_data = json.load(fh)
                fh.seek(0)
                m = hashlib.md5()
                m.update(fh.read().encode("utf8"))
            training_data_hash = m.hexdigest()[:10]
            selector_filename = "askl2_selector_%s_%s_%s_%s.pkl" % (
                autosklearn.__version__,
                sklearn.__version__,
                metric.name,
                training_data_hash,
            )
            selector_directory = os.environ.get("XDG_CACHE_HOME")
            if selector_directory is None:
                selector_directory = pathlib.Path.home()
            selector_directory = (
                pathlib.Path(selector_directory).joinpath("auto-sklearn").expanduser()
            )
            self.selector_files[metric.name] = selector_directory / selector_filename
            metafeatures = pd.DataFrame(training_data["metafeatures"])
            self.strategies = training_data["strategies"]
            y_values = pd.DataFrame(
                training_data["y_values"],
                columns=self.strategies,
                index=metafeatures.index,
            )
            minima_for_methods = training_data["minima_for_methods"]
            maxima_for_methods = training_data["maxima_for_methods"]
            default_strategies = training_data["tie_break_order"]
            if not self.selector_files[metric.name].exists():
                self.required_training = True
                selector = autosklearn.experimental.selector.OVORF(
                    configuration=training_data["configuration"],
                    random_state=np.random.RandomState(1),
                    n_estimators=500,
                    tie_break_order=default_strategies,
                )
                selector = autosklearn.experimental.selector.FallbackWrapper(
                    selector, default_strategies
                )
                selector.fit(
                    X=metafeatures,
                    y=y_values,
                    minima=minima_for_methods,
                    maxima=maxima_for_methods,
                )
                self.selector_files[metric.name].parent.mkdir(
                    exist_ok=True, parents=True
                )

                try:
                    with open(self.selector_files[metric.name], "wb") as fh:
                        pickle.dump(selector, fh)
                except Exception as e:
                    print(
                        "AutoSklearn2Classifier needs to create a selector file under "
                        "the user's home directory or XDG_CACHE_HOME. Nevertheless "
                        "the path {} is not writable.".format(
                            self.selector_files[metric.name]
                        )
                    )
                    raise e

    def fit(
        self,
        X,
        y,
        X_test=None,
        y_test=None,
        metric=None,
        feat_type=None,
        dataset_name=None,
    ):

        # TODO
        # regularly check https://github.com/scikit-learn/scikit-learn/issues/15336
        # whether histogram gradient boosting in scikit-learn finally support
        # sparse data
        is_sparse = scipy.sparse.issparse(X)
        if is_sparse:
            include_estimators = [
                "extra_trees",
                "passive_aggressive",
                "random_forest",
                "sgd",
                "mlp",
            ]
        else:
            include_estimators = [
                "extra_trees",
                "passive_aggressive",
                "random_forest",
                "sgd",
                "gradient_boosting",
                "mlp",
            ]
        self.include["classifier"] = include_estimators

        if self.metric is None:
            if len(y.shape) == 1 or y.shape[1] == 1:
                self.metric = accuracy
            else:
                self.metric = log_loss

        if self.metric in self.selector_metrics:
            metric_name = self.metric.name
            selector_file = self.selector_files[metric_name]
        else:
            metric_name = "balanced_accuracy"
            selector_file = self.selector_files[metric_name]
        with open(selector_file, "rb") as fh:
            selector = pickle.load(fh)

        metafeatures = pd.DataFrame(
            {dataset_name: [X.shape[1], X.shape[0]]}
        ).transpose()
        selection = np.argmax(selector.predict(metafeatures))
        automl_policy = self.strategies[selection]

        setting = {
            "RF_None_holdout_iterative_es_if": {
                "resampling_strategy": "holdout-iterative-fit",
                "fidelity": None,
            },
            "RF_None_3CV_iterative_es_if": {
                "resampling_strategy": "cv-iterative-fit",
                "folds": 3,
                "fidelity": None,
            },
            "RF_None_5CV_iterative_es_if": {
                "resampling_strategy": "cv-iterative-fit",
                "folds": 5,
                "fidelity": None,
            },
            "RF_None_10CV_iterative_es_if": {
                "resampling_strategy": "cv-iterative-fit",
                "folds": 10,
                "fidelity": None,
            },
            "RF_SH-eta4-i_holdout_iterative_es_if": {
                "resampling_strategy": "holdout-iterative-fit",
                "fidelity": "SH",
            },
            "RF_SH-eta4-i_3CV_iterative_es_if": {
                "resampling_strategy": "cv-iterative-fit",
                "folds": 3,
                "fidelity": "SH",
            },
            "RF_SH-eta4-i_5CV_iterative_es_if": {
                "resampling_strategy": "cv-iterative-fit",
                "folds": 5,
                "fidelity": "SH",
            },
            "RF_SH-eta4-i_10CV_iterative_es_if": {
                "resampling_strategy": "cv-iterative-fit",
                "folds": 10,
                "fidelity": "SH",
            },
        }[automl_policy]

        resampling_strategy = setting["resampling_strategy"]
        if resampling_strategy == "cv-iterative-fit":
            resampling_strategy_kwargs = {"folds": setting["folds"]}
        else:
            resampling_strategy_kwargs = None

        portfolio_file = (
            self.this_directory
            / metric_name
            / "askl2_portfolios"
            / ("%s.json" % automl_policy)
        )
        with open(portfolio_file) as fh:
            portfolio_json = json.load(fh)
        portfolio = portfolio_json["portfolio"]

        if setting["fidelity"] == "SH":
            smac_callback = SHObjectCallback("iterations", 4, 5.0, portfolio)
        else:
            smac_callback = SmacObjectCallback(portfolio)

        self.resampling_strategy = resampling_strategy
        self.resampling_strategy_arguments = resampling_strategy_kwargs
        self.get_smac_object_callback = smac_callback
        return super().fit(
            X=X,
            y=y,
            X_test=X_test,
            y_test=y_test,
            feat_type=feat_type,
            dataset_name=dataset_name,
        )



import numpy as np
import pandas as pd
import sklearn.datasets
import sklearn.model_selection
from pandas.api.types import is_categorical_dtype, is_numeric_dtype, is_string_dtype
from scipy import sparse

from autosklearn.data.feature_validator import FeatureValidator

import pytest


# Fixtures to be used in this class. By default all elements have 100 datapoints
@pytest.fixture
def input_data_featuretest(request):
    if request.param == "numpy_categoricalonly_nonan":
        return np.random.randint(10, size=(100, 10))
    elif request.param == "numpy_numericalonly_nonan":
        return np.random.uniform(10, size=(100, 10))
    elif request.param == "numpy_mixed_nonan":
        return np.column_stack(
            [
                np.random.uniform(10, size=(100, 3)),
                np.random.randint(10, size=(100, 3)),
                np.random.uniform(10, size=(100, 3)),
                np.random.randint(10, size=(100, 1)),
            ]
        )
    elif request.param == "numpy_string_nonan":
        return np.array(
            [
                ["a", "b", "c", "a", "b", "c"],
                ["a", "b", "d", "r", "b", "c"],
            ]
        )
    elif request.param == "numpy_categoricalonly_nan":
        array = np.random.randint(10, size=(100, 10)).astype("float")
        array[50, 0:5] = np.nan
        return array
    elif request.param == "numpy_numericalonly_nan":
        array = np.random.uniform(10, size=(100, 10)).astype("float")
        array[50, 0:5] = np.nan
        # Somehow array is changed to dtype object after np.nan
        return array.astype("float")
    elif request.param == "numpy_mixed_nan":
        array = np.column_stack(
            [
                np.random.uniform(10, size=(100, 3)),
                np.random.randint(10, size=(100, 3)),
                np.random.uniform(10, size=(100, 3)),
                np.random.randint(10, size=(100, 1)),
            ]
        )
        array[50, 0:5] = np.nan
        return array
    elif request.param == "numpy_string_nan":
        return np.array(
            [
                ["a", "b", "c", "a", "b", "c"],
                [np.nan, "b", "d", "r", "b", "c"],
            ]
        )
    elif request.param == "pandas_categoricalonly_nonan":
        return pd.DataFrame(
            [
                {"A": 1, "B": 2},
                {"A": 3, "B": 4},
            ],
            dtype="category",
        )
    elif request.param == "pandas_numericalonly_nonan":
        return pd.DataFrame(
            [
                {"A": 1, "B": 2},
                {"A": 3, "B": 4},
            ],
            dtype="float",
        )
    elif request.param == "pandas_mixed_nonan":
        frame = pd.DataFrame(
            [
                {"A": 1, "B": 2},
                {"A": 3, "B": 4},
            ],
            dtype="category",
        )
        frame["B"] = pd.to_numeric(frame["B"])
        return frame
    elif request.param == "pandas_categoricalonly_nan":
        return pd.DataFrame(
            [
                {"A": 1, "B": 2, "C": np.nan},
                {"A": 3, "C": np.nan},
            ],
            dtype="category",
        )
    elif request.param == "pandas_numericalonly_nan":
        return pd.DataFrame(
            [
                {"A": 1, "B": 2, "C": np.nan},
                {"A": 3, "C": np.nan},
            ],
            dtype="float",
        )
    elif request.param == "pandas_mixed_nan":
        frame = pd.DataFrame(
            [
                {"A": 1, "B": 2, "C": 8},
                {"A": 3, "B": 4},
            ],
            dtype="category",
        )
        frame["B"] = pd.to_numeric(frame["B"])
        return frame
    elif request.param == "pandas_string_nonan":
        return pd.DataFrame(
            [
                {"A": 1, "B": 2},
                {"A": 3, "B": 4},
            ],
            dtype="string",
        )
    elif request.param == "list_categoricalonly_nonan":
        return [
            ["a", "b", "c", "d"],
            ["e", "f", "c", "d"],
        ]
    elif request.param == "list_numericalonly_nonan":
        return [[1, 2, 3, 4], [5, 6, 7, 8]]
    elif request.param == "list_mixed_nonan":
        return [["a", 2, 3, 4], ["b", 6, 7, 8]]
    elif request.param == "list_categoricalonly_nan":
        return [
            ["a", "b", "c", np.nan],
            ["e", "f", "c", "d"],
        ]
    elif request.param == "list_numericalonly_nan":
        return [[1, 2, 3, np.nan], [5, 6, 7, 8]]
    elif request.param == "list_mixed_nan":
        return [["a", np.nan, 3, 4], ["b", 6, 7, 8]]
    elif "sparse" in request.param:
        # We expect the names to be of the type sparse_csc_nonan
        sparse_, type_, nan_ = request.param.split("_")
        if "nonan" in nan_:
            data = np.ones(3)
        else:
            data = np.array([1, 2, np.nan])

        # Then the type of sparse
        row_ind = np.array([0, 1, 2])
        col_ind = np.array([1, 2, 1])
        if "csc" in type_:
            return sparse.csc_matrix((data, (row_ind, col_ind)))
        elif "csr" in type_:
            return sparse.csr_matrix((data, (row_ind, col_ind)))
        elif "coo" in type_:
            return sparse.coo_matrix((data, (row_ind, col_ind)))
        elif "bsr" in type_:
            return sparse.bsr_matrix((data, (row_ind, col_ind)))
        elif "lil" in type_:
            return sparse.lil_matrix((data))
        elif "dok" in type_:
            return sparse.dok_matrix(np.vstack((data, data, data)))
        elif "dia" in type_:
            return sparse.dia_matrix(np.vstack((data, data, data)))
        else:
            ValueError("Unsupported indirect fixture {}".format(request.param))
    elif "openml" in request.param:
        _, openml_id = request.param.split("_")
        X, y = sklearn.datasets.fetch_openml(
            data_id=int(openml_id), return_X_y=True, as_frame=True
        )
        return X
    else:
        ValueError("Unsupported indirect fixture {}".format(request.param))


# Actual checks for the features
@pytest.mark.parametrize(
    "input_data_featuretest",
    (
        "numpy_categoricalonly_nonan",
        "numpy_numericalonly_nonan",
        "numpy_mixed_nonan",
        "numpy_categoricalonly_nan",
        "numpy_numericalonly_nan",
        "numpy_mixed_nan",
        "pandas_categoricalonly_nonan",
        "pandas_numericalonly_nonan",
        "pandas_mixed_nonan",
        "pandas_numericalonly_nan",
        "list_numericalonly_nonan",
        "list_numericalonly_nan",
        "sparse_bsr_nonan",
        "sparse_bsr_nan",
        "sparse_coo_nonan",
        "sparse_coo_nan",
        "sparse_csc_nonan",
        "sparse_csc_nan",
        "sparse_csr_nonan",
        "sparse_csr_nan",
        "sparse_dia_nonan",
        "sparse_dia_nan",
        "sparse_dok_nonan",
        "sparse_dok_nan",
        "sparse_lil_nonan",
        "sparse_lil_nan",
        "openml_40981",  # Australian
    ),
    indirect=True,
)
def test_featurevalidator_supported_types(input_data_featuretest):
    validator = FeatureValidator()
    validator.fit(input_data_featuretest, input_data_featuretest)
    transformed_X = validator.transform(input_data_featuretest)
    if sparse.issparse(input_data_featuretest):
        assert sparse.issparse(transformed_X)
    elif isinstance(input_data_featuretest, list):
        assert isinstance(transformed_X, pd.DataFrame)
    else:
        assert isinstance(transformed_X, type(input_data_featuretest))
    assert np.shape(input_data_featuretest) == np.shape(transformed_X)
    assert validator._is_fitted


@pytest.mark.parametrize(
    "input_data_featuretest",
    (
        "numpy_string_nonan",
        "numpy_string_nan",
    ),
    indirect=True,
)
def test_featurevalidator_unsupported_numpy(input_data_featuretest):
    validator = FeatureValidator()
    with pytest.raises(
        ValueError, match=r".*When providing a numpy array.*not supported."
    ):
        validator.fit(input_data_featuretest)


@pytest.mark.parametrize(
    "input_data_featuretest",
    (
        "numpy_categoricalonly_nonan",
        "numpy_mixed_nonan",
        "numpy_categoricalonly_nan",
        "numpy_mixed_nan",
        "pandas_categoricalonly_nonan",
        "pandas_mixed_nonan",
        "sparse_bsr_nonan",
        "sparse_bsr_nan",
        "sparse_coo_nonan",
        "sparse_coo_nan",
        "sparse_csc_nonan",
        "sparse_csc_nan",
        "sparse_csr_nonan",
        "sparse_csr_nan",
        "sparse_dia_nonan",
        "sparse_dia_nan",
        "sparse_dok_nonan",
        "sparse_dok_nan",
        "sparse_lil_nonan",
    ),
    indirect=True,
)
def test_featurevalidator_fitontypeA_transformtypeB(input_data_featuretest):
    """
    Check if we can fit in a given type (numpy) yet transform
    if the user changes the type (pandas then)
    """
    validator = FeatureValidator()
    validator.fit(input_data_featuretest, input_data_featuretest)
    if isinstance(input_data_featuretest, pd.DataFrame):
        complementary_type = input_data_featuretest.to_numpy()
    elif isinstance(input_data_featuretest, np.ndarray):
        complementary_type = pd.DataFrame(input_data_featuretest)
    elif isinstance(input_data_featuretest, list):
        complementary_type = pd.DataFrame(input_data_featuretest)
    elif sparse.issparse(input_data_featuretest):
        complementary_type = sparse.csr_matrix(input_data_featuretest.todense())
    else:
        raise ValueError(type(input_data_featuretest))
    transformed_X = validator.transform(complementary_type)
    assert np.shape(input_data_featuretest) == np.shape(transformed_X)
    assert validator._is_fitted


def test_featurevalidatorget_feat_type_from_columns():
    """
    Makes sure that encoded columns are returned by get_feat_type_from_columns
    whereas numerical columns are not returned
    """
    validator = FeatureValidator()

    df = pd.DataFrame(
        [
            {"int": 1, "float": 1.0, "category": "one", "bool": True},
            {"int": 2, "float": 2.0, "category": "two", "bool": False},
        ]
    )

    for col in df.columns:
        df[col] = df[col].astype(col)

    feature_types = validator.get_feat_type_from_columns(df)

    assert feature_types == {
        "int": "numerical",
        "float": "numerical",
        "category": "categorical",
        "bool": "categorical",
    }


def test_features_unsupported_calls_are_raised():
    """
    Makes sure we raise a proper message to the user,
    when providing not supported data input or using the validator in a way that is not
    expected
    """
    validator = FeatureValidator()
    with pytest.raises(ValueError, match=r"Auto-sklearn does not support time"):
        validator.fit(pd.DataFrame({"datetime": [pd.Timestamp("20180310")]}))
    with pytest.raises(
        ValueError, match=r"Auto-sklearn only supports.*yet, the provided input"
    ):
        validator.fit({"input1": 1, "input2": 2})
    validator = FeatureValidator()
    with pytest.raises(
        ValueError, match=r"The feature dimensionality of the train and test"
    ):
        validator.fit(
            X_train=np.array([[1, 2, 3], [4, 5, 6]]),
            X_test=np.array([[1, 2, 3, 4], [4, 5, 6, 7]]),
        )
    with pytest.raises(
        ValueError, match=r"Cannot call transform on a validator that is not fit"
    ):
        validator.transform(np.array([[1, 2, 3], [4, 5, 6]]))
    validator = FeatureValidator(feat_type=["Numerical"])
    with pytest.raises(
        ValueError, match=r"providing the option feat_type to the fit method is.*"
    ):
        validator.fit(pd.DataFrame([[1, 2, 3], [4, 5, 6]]))
    with pytest.raises(ValueError, match=r"feat_type does not have same number of.*"):
        validator.fit(np.array([[1, 2, 3], [4, 5, 6]]))
    validator = FeatureValidator(feat_type=[1, 2, 3])
    with pytest.raises(ValueError, match=r"feat_type must only contain strings.*"):
        validator.fit(np.array([[1, 2, 3], [4, 5, 6]]))
    validator = FeatureValidator(feat_type=["1", "2", "3"])
    with pytest.raises(
        ValueError, match=r"Only `Categorical`, `Numerical` and `String` are.*"
    ):
        validator.fit(np.array([[1, 2, 3], [4, 5, 6]]))


def test_no_new_category_after_fit():
    """
    This test makes sure that we can actually pass new categories to the estimator
    without throwing an error
    """
    # Then make sure we catch categorical extra categories
    x = pd.DataFrame({"A": [1, 2, 3, 4], "B": [5, 6, 7, 8]}, dtype="category")
    validator = FeatureValidator()
    validator.fit(x)
    x["A"] = x["A"].apply(lambda x: x * x)
    validator.transform(x)


# Actual checks for the features
@pytest.mark.parametrize(
    "openml_id",
    (
        40981,  # Australian
        3,  # kr-vs-kp
        1468,  # cnae-9
        40975,  # car
        40984,  # Segment
    ),
)
@pytest.mark.parametrize("train_data_type", ("numpy", "pandas", "list"))
@pytest.mark.parametrize("test_data_type", ("numpy", "pandas", "list"))
def test_featurevalidator_new_data_after_fit(
    openml_id, train_data_type, test_data_type
):
    # List is currently not supported as infer_objects
    # cast list objects to type objects
    if train_data_type == "list" or test_data_type == "list":
        pytest.skip()

    validator = FeatureValidator()

    if train_data_type == "numpy":
        X, y = sklearn.datasets.fetch_openml(
            data_id=openml_id, return_X_y=True, as_frame=False
        )
    elif train_data_type == "pandas":
        X, y = sklearn.datasets.fetch_openml(
            data_id=openml_id, return_X_y=True, as_frame=True
        )
    else:
        X, y = sklearn.datasets.fetch_openml(
            data_id=openml_id, return_X_y=True, as_frame=True
        )
        X = X.values.tolist()
        y = y.values.tolist()

    X_train, X_test, y_train, y_test = sklearn.model_selection.train_test_split(
        X, y, random_state=1
    )

    validator.fit(X_train)

    transformed_X = validator.transform(X_test)

    # Basic Checking
    if sparse.issparse(input_data_featuretest):
        assert sparse.issparse(transformed_X)
    elif isinstance(input_data_featuretest, list):
        assert isinstance(transformed_X, pd.DataFrame)
    else:
        assert isinstance(transformed_X, type(X_train))
    assert np.shape(X_test) == np.shape(transformed_X)


@pytest.mark.parametrize(
    "openml_id",
    (
        40981,  # Australian
        3,  # kr-vs-kp
        1468,  # cnae-9
        40975,  # car
        40984,  # Segment
        2,  # anneal
    ),
)
def test_list_to_dataframe(openml_id):
    X_pandas, y_pandas = sklearn.datasets.fetch_openml(
        data_id=openml_id, return_X_y=True, as_frame=True
    )
    X_train, X_test, y_train, y_test = sklearn.model_selection.train_test_split(
        X_pandas, y_pandas, random_state=1
    )

    X_list = X_train.values.tolist()
    validator = FeatureValidator()
    validator.fit(X_list)
    transformed_X = validator.transform(X_list)
    for i, col in enumerate(X_pandas.columns):
        if is_numeric_dtype(X_pandas[col].dtype):
            # convert dtype translates 72.0 to 72. Be robust against this!
            assert is_numeric_dtype(transformed_X[i].dtype)
        elif is_string_dtype(X_pandas[col].dtype):
            assert is_string_dtype(X_pandas[col].dtype)
        elif is_categorical_dtype(X_pandas[col].dtype):
            assert is_categorical_dtype(X_pandas[col].dtype)
        else:
            raise NotImplementedError

    # Also make sure that at testing time
    # this work
    transformed_X = validator.transform(X_test.values.tolist())
    for i, col in enumerate(X_pandas.columns):
        if is_numeric_dtype(X_pandas[col].dtype):
            # convert dtype translates 72.0 to 72. Be robust against this!
            assert is_numeric_dtype(transformed_X[i].dtype)
        elif is_string_dtype(X_pandas[col].dtype):
            assert is_string_dtype(X_pandas[col].dtype)
        elif is_categorical_dtype(X_pandas[col].dtype):
            assert is_categorical_dtype(X_pandas[col].dtype)
        else:
            raise NotImplementedError


@pytest.mark.parametrize(
    "input_data_featuretest",
    (
        "sparse_bsr_nonan",
        "sparse_bsr_nan",
        "sparse_coo_nonan",
        "sparse_coo_nan",
        "sparse_csc_nonan",
        "sparse_csc_nan",
        "sparse_csr_nonan",
        "sparse_csr_nan",
        "sparse_dia_nonan",
        "sparse_dia_nan",
        "sparse_dok_nonan",
        "sparse_dok_nan",
        "sparse_lil_nonan",
        "sparse_lil_nan",
    ),
    indirect=True,
)
def test_sparse_output_is_csr(input_data_featuretest):
    validator = FeatureValidator()
    validator.fit(input_data_featuretest, input_data_featuretest)
    transformed_X = validator.transform(input_data_featuretest)
    assert sparse.issparse(transformed_X)
    assert isinstance(transformed_X, sparse.csr_matrix)


def test_unsupported_dataframe_sparse():
    df = pd.DataFrame({"A": pd.Series(pd.arrays.SparseArray(np.random.randn(10)))})
    validator = FeatureValidator()
    with pytest.raises(
        ValueError, match=r"Auto-sklearn does not yet support sparse pandas"
    ):
        validator.fit(df)


def test_object_columns():
    class Dummy:
        def __init__(self, x):
            self.x = x

        def __call__(self):
            print(self.x)

        def dummy_func(self):
            for i in range(100):
                print("do something 100 times")

    dummy_object = Dummy(1)
    lst = [1, 2, 3]
    array = np.array([1, 2, 3])
    dummy_string = "dummy string"

    df = pd.DataFrame(
        {
            "dummy_object": [dummy_object] * 4,
            "dummy_lst": [lst] * 4,
            "dummy_array": [array] * 4,
            "dummy_string": [dummy_string] * 4,
            "type_mix_column": [dummy_string, dummy_object, array, lst],
            "cat_column": ["a", "b", "a", "b"],
        }
    )
    df["cat_column"] = df["cat_column"].astype("category")

    with pytest.warns(
        UserWarning,
        match=r"Input Column dummy_object has "
        r"generic type object. "
        r"Autosklearn will treat "
        r"this column as string. "
        r"Please ensure that this setting "
        r"is suitable for your task.",
    ):
        validator = FeatureValidator()
        feat_type = validator.get_feat_type_from_columns(df)

    column_types = {
        "dummy_object": "string",
        "dummy_lst": "string",
        "dummy_array": "string",
        "dummy_string": "string",
        "type_mix_column": "string",
        "cat_column": "categorical",
    }

    assert feat_type == column_types


def test_allow_string_feature():
    df = pd.DataFrame({"Text": ["Hello", "how are you?"]})
    with pytest.warns(
        UserWarning,
        match=r"Input Column Text has generic type object. "
        r"Autosklearn will treat this column as string. "
        r"Please ensure that this setting is suitable for your task.",
    ):
        validator = FeatureValidator(allow_string_features=False)
        feat_type = validator.get_feat_type_from_columns(df)

    column_types = {"Text": "categorical"}
    assert feat_type == column_types

    df["Text"] = df["Text"].astype("string")
    with pytest.warns(
        UserWarning,
        match=r"you disabled text encoding column Text will be " r"encoded as category",
    ):
        validator = FeatureValidator(allow_string_features=False)
        feat_type = validator.get_feat_type_from_columns(df)

    column_types = {"Text": "categorical"}
    assert feat_type == column_types



from typing import Optional

from ConfigSpace.conditions import EqualsCondition, InCondition
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import (
    DENSE,
    INPUT,
    SIGNED_DATA,
    SPARSE,
    UNSIGNED_DATA,
)


class Nystroem(AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self, kernel, n_components, gamma=1.0, degree=3, coef0=1, random_state=None
    ):
        self.kernel = kernel
        self.n_components = n_components
        self.gamma = gamma
        self.degree = degree
        self.coef0 = coef0
        self.random_state = random_state

    def fit(self, X, Y=None):
        import scipy.sparse
        import sklearn.kernel_approximation

        self.n_components = int(self.n_components)
        self.gamma = float(self.gamma)
        self.degree = int(self.degree)
        self.coef0 = float(self.coef0)

        self.preprocessor = sklearn.kernel_approximation.Nystroem(
            kernel=self.kernel,
            n_components=self.n_components,
            gamma=self.gamma,
            degree=self.degree,
            coef0=self.coef0,
            random_state=self.random_state,
        )

        # Because the pipeline guarantees that each feature is positive,
        # clip all values below zero to zero
        if self.kernel == "chi2":
            if scipy.sparse.issparse(X):
                X.data[X.data < 0] = 0.0
            else:
                X[X < 0] = 0.0

        self.preprocessor.fit(X)
        return self

    def transform(self, X):
        import scipy.sparse

        # Because the pipeline guarantees that each feature is positive,
        # clip all values below zero to zero
        if self.kernel == "chi2":
            if scipy.sparse.issparse(X):
                X.data[X.data < 0] = 0.0
            else:
                X[X < 0] = 0.0

        if self.preprocessor is None:
            raise NotImplementedError()
        return self.preprocessor.transform(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        data_type = UNSIGNED_DATA

        if dataset_properties is not None:
            signed = dataset_properties.get("signed")
            if signed is not None:
                data_type = SIGNED_DATA if signed is True else UNSIGNED_DATA
        return {
            "shortname": "Nystroem",
            "name": "Nystroem kernel approximation",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "is_deterministic": True,
            "input": (SPARSE, DENSE, data_type),
            "output": (INPUT, UNSIGNED_DATA),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        if dataset_properties is not None and (
            dataset_properties.get("sparse") is True
            or dataset_properties.get("signed") is False
        ):
            allow_chi2 = False
        else:
            allow_chi2 = True

        possible_kernels = ["poly", "rbf", "sigmoid", "cosine"]
        if allow_chi2:
            possible_kernels.append("chi2")
        kernel = CategoricalHyperparameter("kernel", possible_kernels, "rbf")
        n_components = UniformIntegerHyperparameter(
            "n_components", 50, 10000, default_value=100, log=True
        )
        gamma = UniformFloatHyperparameter(
            "gamma", 3.0517578125e-05, 8, log=True, default_value=0.1
        )
        degree = UniformIntegerHyperparameter("degree", 2, 5, 3)
        coef0 = UniformFloatHyperparameter("coef0", -1, 1, default_value=0)

        cs = ConfigurationSpace()
        cs.add_hyperparameters([kernel, degree, gamma, coef0, n_components])

        degree_depends_on_poly = EqualsCondition(degree, kernel, "poly")
        coef0_condition = InCondition(coef0, kernel, ["poly", "sigmoid"])

        gamma_kernels = ["poly", "rbf", "sigmoid"]
        if allow_chi2:
            gamma_kernels.append("chi2")
        gamma_condition = InCondition(gamma, kernel, gamma_kernels)
        cs.add_conditions([degree_depends_on_poly, coef0_condition, gamma_condition])
        return cs



import sklearn.ensemble

from autosklearn.pipeline.components.regression.adaboost import AdaboostRegressor

from .test_base import BaseRegressionComponentTest


class AdaBoostComponentTest(BaseRegressionComponentTest):

    __test__ = True

    res = dict()
    res["default_boston"] = 0.5951486466070626
    res["default_boston_iterative"] = None
    res["default_boston_sparse"] = 0.18067558132702222
    res["default_boston_iterative_sparse"] = None
    res["default_diabetes"] = 0.250565253614339
    res["default_diabetes_iterative"] = None
    res["default_diabetes_sparse"] = 0.09126705185668416
    res["default_diabetes_iterative_sparse"] = None

    sk_mod = sklearn.ensemble.AdaBoostRegressor

    module = AdaboostRegressor



import numpy as np

from autosklearn.pipeline.components.feature_preprocessing.no_preprocessing import (
    NoPreprocessing,
)
from autosklearn.pipeline.util import PreprocessingTestCase, _test_preprocessing


class NoneComponentTest(PreprocessingTestCase):
    def test_default_configuration(self):
        transformation, original = _test_preprocessing(NoPreprocessing)
        self.assertEqual(transformation.shape[0], original.shape[0])
        self.assertEqual(transformation.shape[1], original.shape[1])
        self.assertFalse((transformation == 0).all())
        self.assertEqual(np.sum(original), np.sum(transformation))
        self.assertEqual(np.min(original), np.min(transformation))
        self.assertEqual(np.max(original), np.max(transformation))
        self.assertEqual(np.std(original), np.std(transformation))
        self.assertEqual(np.mean(original), np.mean(transformation))

    def test_preprocessing_dtype(self):
        super(NoneComponentTest, self)._test_preprocessing_dtype(NoPreprocessing)



from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from ConfigSpace.configuration_space import Configuration, ConfigurationSpace
from sklearn.base import BaseEstimator

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, BasePipeline
from autosklearn.pipeline.components.data_preprocessing.categorical_encoding import (  # noqa: E501
    OHEChoice,
)
from autosklearn.pipeline.components.data_preprocessing.categorical_encoding.encoding import (  # noqa: E501
    OrdinalEncoding,
)
from autosklearn.pipeline.components.data_preprocessing.category_shift.category_shift import (  # noqa: E501
    CategoryShift,
)
from autosklearn.pipeline.components.data_preprocessing.imputation.categorical_imputation import (  # noqa: E501
    CategoricalImputation,
)
from autosklearn.pipeline.components.data_preprocessing.minority_coalescense import (
    CoalescenseChoice,
)
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA


class CategoricalPreprocessingPipeline(BasePipeline):
    """This class implements a pipeline for data preprocessing of categorical features.
    It assumes that the data to be transformed is made only of categorical features.
    The steps of this pipeline are:
        1 - Category shift: Adds 3 to every category value
        2 - Imputation: Assign category 2 to missing values (NaN).
        3 - Minority coalescence: Assign category 1 to all categories whose occurrence
            don't sum-up to a certain minimum fraction
        4 - One hot encoding: usual sklearn one hot encoding
    Parameters
    ----------
    config : ConfigSpace.configuration_space.Configuration
        The configuration to evaluate.
    random_state : Optional[int | RandomState]
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance
        used by `np.random`."""

    def __init__(
        self,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        config: Optional[Configuration] = None,
        steps: Optional[List[Tuple[str, BaseEstimator]]] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
        include: Optional[Dict[str, str]] = None,
        exclude: Optional[Dict[str, str]] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        init_params: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._output_dtype = np.int32
        super().__init__(
            config=config,
            steps=steps,
            dataset_properties=dataset_properties,
            include=include,
            exclude=exclude,
            random_state=random_state,
            init_params=init_params,
            feat_type=feat_type,
        )

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "cat_datapreproc",
            "name": "categorical data preprocessing",
            "handles_missing_values": True,
            "handles_nominal_values": True,
            "handles_numerical_features": True,
            "prefers_data_scaled": False,
            "prefers_data_normalized": False,
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "is_deterministic": True,
            # TODO find out if this is right!
            "handles_sparse": True,
            "handles_dense": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
            "preferred_dtype": None,
        }

    def _get_hyperparameter_search_space(
        self,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        include: Optional[Dict[str, str]] = None,
        exclude: Optional[Dict[str, str]] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> ConfigurationSpace:
        """Create the hyperparameter configuration space.
        Returns
        -------
        cs : ConfigSpace.configuration_space.Configuration
            The configuration space describing the SimpleRegressionClassifier.
        """
        cs = ConfigurationSpace()

        if dataset_properties is None or not isinstance(dataset_properties, dict):
            dataset_properties = dict()

        cs = self._get_base_search_space(
            cs=cs,
            feat_type=feat_type,
            dataset_properties=dataset_properties,
            exclude=exclude,
            include=include,
            pipeline=self.steps,
        )

        return cs

    def _get_pipeline_steps(
        self,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[Dict[str, str]] = None,
    ) -> List[Tuple[str, BaseEstimator]]:
        steps = []

        default_dataset_properties = {}
        if dataset_properties is not None and isinstance(dataset_properties, dict):
            default_dataset_properties.update(dataset_properties)

        steps = [
            ("imputation", CategoricalImputation(random_state=self.random_state)),
            ("encoding", OrdinalEncoding(random_state=self.random_state)),
            ("category_shift", CategoryShift(random_state=self.random_state)),
            (
                "category_coalescence",
                CoalescenseChoice(
                    feat_type=feat_type,
                    dataset_properties=default_dataset_properties,
                    random_state=self.random_state,
                ),
            ),
            (
                "categorical_encoding",
                OHEChoice(
                    feat_type=feat_type,
                    dataset_properties=default_dataset_properties,
                    random_state=self.random_state,
                ),
            ),
        ]

        return steps

    def _get_estimator_hyperparameter_name(self) -> str:
        return "categorical data preprocessing"



"""Test specific ways of calling `fit` of AutoML"""
from typing import Any, Callable, Dict, Optional, Tuple, Union

import numpy as np
from dask.distributed import Client
from smac.facade.roar_facade import ROAR
from smac.scenario.scenario import Scenario

from autosklearn.automl import AutoML
from autosklearn.constants import MULTICLASS_CLASSIFICATION

from pytest_cases import parametrize


@parametrize("dataset, task, bounds", [("iris", MULTICLASS_CLASSIFICATION, (0.8, 1.0))])
def test_fit_roar(
    dataset: str,
    task: int,
    bounds: Tuple[float, float],
    dask_client_single_worker: Client,
    make_automl: Callable[..., AutoML],
    make_sklearn_dataset: Callable[..., Tuple[np.ndarray, ...]],
) -> None:
    """
    Parameters
    ----------
    dataset : str
        The name of the dataset

    task : int
        The task type of the dataset

    bounds : Tuple[float, float]
        The bounds the final score should be in, (lowest, upper)

    Fixtures
    --------
    make_automl : Callable[..., AutoML]
        Factory for making an AutoML instance

    make_sklearn_dataset : Callable[..., Tuple[np.ndarray, ...]]
        Factory for getting a dataset

    Expects
    -------
    * Should fit without a problem using a different smac object
    """

    def get_roar_object_callback(
        scenario_dict: Dict,
        seed: Optional[Union[int, np.random.RandomState]],
        ta: Callable,
        ta_kwargs: Dict,
        dask_client: Client,
        n_jobs: int,
        **kwargs: Any,
    ) -> ROAR:
        """Random online adaptive racing.

        http://ml.informatik.uni-freiburg.de/papers/11-LION5-SMAC.pdf
        """
        scenario = Scenario(scenario_dict)
        return ROAR(
            run_id=seed,
            scenario=scenario,
            rng=seed,
            tae_runner=ta,
            tae_runner_kwargs=ta_kwargs,
            dask_client=dask_client,
            n_jobs=n_jobs,
        )

    X_train, Y_train, X_test, Y_test = make_sklearn_dataset(dataset)
    automl = make_automl(
        initial_configurations_via_metalearning=0,
        get_smac_object_callback=get_roar_object_callback,
        dask_client=dask_client_single_worker,
    )

    automl.fit(X_train, Y_train, task=task)

    score = automl.score(X_test, Y_test)
    assert score > 0.8



from typing import Optional

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    Constant,
    UniformIntegerHyperparameter,
    UnParametrizedHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, SIGNED_DATA, SPARSE, UNSIGNED_DATA
from autosklearn.util.common import check_for_bool, check_none


class RandomTreesEmbedding(AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self,
        n_estimators,
        max_depth,
        min_samples_split,
        min_samples_leaf,
        min_weight_fraction_leaf,
        max_leaf_nodes,
        bootstrap,
        sparse_output=True,
        n_jobs=1,
        random_state=None,
    ):
        self.n_estimators = n_estimators
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_leaf_nodes = max_leaf_nodes
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.bootstrap = bootstrap
        self.sparse_output = sparse_output
        self.n_jobs = n_jobs
        self.random_state = random_state

    def _fit(self, X, Y=None):
        import sklearn.ensemble

        self.n_estimators = int(self.n_estimators)
        if check_none(self.max_depth):
            self.max_depth = None
        else:
            self.max_depth = int(self.max_depth)
        self.min_samples_split = int(self.min_samples_split)
        self.min_samples_leaf = int(self.min_samples_leaf)
        if check_none(self.max_leaf_nodes):
            self.max_leaf_nodes = None
        else:
            self.max_leaf_nodes = int(self.max_leaf_nodes)
        self.min_weight_fraction_leaf = float(self.min_weight_fraction_leaf)
        self.bootstrap = check_for_bool(self.bootstrap)

        self.preprocessor = sklearn.ensemble.RandomTreesEmbedding(
            n_estimators=self.n_estimators,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            max_leaf_nodes=self.max_leaf_nodes,
            sparse_output=self.sparse_output,
            n_jobs=self.n_jobs,
            random_state=self.random_state,
        )
        self.preprocessor.fit(X, Y)
        return self

    def fit(self, X, y):
        self._fit(X)
        return self

    def fit_transform(self, X, y=None):
        return self._fit(X)

    def transform(self, X):
        if self.preprocessor is None:
            raise NotImplementedError()
        return self.preprocessor.transform(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "RandomTreesEmbedding",
            "name": "Random Trees Embedding",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (SPARSE, SIGNED_DATA),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        n_estimators = UniformIntegerHyperparameter(
            name="n_estimators", lower=10, upper=100, default_value=10
        )
        max_depth = UniformIntegerHyperparameter(
            name="max_depth", lower=2, upper=10, default_value=5
        )
        min_samples_split = UniformIntegerHyperparameter(
            name="min_samples_split", lower=2, upper=20, default_value=2
        )
        min_samples_leaf = UniformIntegerHyperparameter(
            name="min_samples_leaf", lower=1, upper=20, default_value=1
        )
        min_weight_fraction_leaf = Constant("min_weight_fraction_leaf", 1.0)
        max_leaf_nodes = UnParametrizedHyperparameter(
            name="max_leaf_nodes", value="None"
        )
        bootstrap = CategoricalHyperparameter("bootstrap", ["True", "False"])
        cs = ConfigurationSpace()
        cs.add_hyperparameters(
            [
                n_estimators,
                max_depth,
                min_samples_split,
                min_samples_leaf,
                min_weight_fraction_leaf,
                max_leaf_nodes,
                bootstrap,
            ]
        )
        return cs



from __future__ import annotations

from typing import Callable

import os
from distutils.dir_util import copy_tree
from pathlib import Path

from autosklearn.automl_common.common.utils.backend import Backend, create
from autosklearn.data.xy_data_manager import XYDataManager

from pytest import fixture

HERE = Path(__file__).parent.resolve()
DATAPATH = HERE.parent / "data"


def copy_backend(old: Backend | Path | str, new: Backend | Path | str) -> Backend:
    """Transfers a backend to a new path

    Parameters
    ----------
    old_backend: Backend | Path | str
        The backend to transfer from

    new_path: Backend | Path | str
        Where to place the new backend

    Returns
    -------
    Backend
        The new backend with the contents of the old
    """
    if isinstance(new, str):
        new_backend = create(
            temporary_directory=new,
            output_directory=None,
            prefix="auto-sklearn",
        )
    elif isinstance(new, Path):
        new_backend = create(
            temporary_directory=str(new),
            output_directory=None,
            prefix="auto-sklearn",
        )
    else:
        new_backend = new

    dst = new_backend.temporary_directory

    if isinstance(old, str):
        src = old
    elif isinstance(old, Path):
        src = str(old)
    else:
        src = old.temporary_directory

    copy_tree(src, dst)

    return new_backend


# TODO Update to return path once everything can use a path
@fixture
def tmp_dir(tmp_path: Path) -> str:
    """
    Fixtures
    --------
    tmp_path : Path
        Built in pytest fixture

    Returns
    -------
    str
        The directory as a str
    """
    return str(tmp_path)


@fixture
def make_backend(tmp_path: Path) -> Callable[..., Backend]:
    """Make a backend

    Parameters
    ----------
    path: Union[str, Path]
        The path to place the backend at

    template: Optional[Path] = None
        Setup with a pre-existing layout if not None

    Returns
    -------
    Backend
        The created backend object
    """
    # TODO redo once things use paths
    def _make(
        path: str | Path | None = None,
        template: Path | Backend | None = None,
        datamanager: XYDataManager | None = None,
    ) -> Backend:
        if template is not None and datamanager is not None:
            raise ValueError("Does not support template and datamanager")

        if path is None:
            _path = Path(tmp_path) / "backend"
        elif isinstance(path, str):
            _path = Path(path)
        else:
            _path = path

        assert not _path.exists(), "Path exists, Try passing path / 'backend'"

        if template is not None:
            backend = copy_backend(old=template, new=_path)
        else:
            backend = create(
                temporary_directory=str(_path),
                output_directory=None,
                prefix="auto-sklearn",
            )

            if datamanager is not None:
                backend.save_datamanager(datamanager)

        return backend

    return _make


@fixture(scope="function")
def backend(tmp_dir: str, make_backend: Callable) -> Backend:
    """A backend object

    Fixtures
    --------
    tmp_dir : str
        A directory to place the backend at

    make_backend : Callable
        Factory to make a backend

    Returns
    -------
    Backend
        A backend object
    """
    backend_path = os.path.join(tmp_dir, "backend")
    return make_backend(path=backend_path)



import sklearn.neural_network

from autosklearn.pipeline.components.classification.mlp import MLPClassifier

from .test_base import BaseClassificationComponentTest


class MLPComponentTest(BaseClassificationComponentTest):
    # NOTE: `default_iris_proba_places`
    #
    # Github runners seem to indeterministicly fail `test_default_iris_proba`
    # meaning 'default_irish_proba_places' needs to be set.
    # There are known differences to occur on different platforms.
    # https://github.com/scikit-learn/scikit-learn/issues/13108#issuecomment-461696681
    #
    # We are assuming results are deterministic on a given platform as locally
    # there is no randomness i.e. performing the same test 100 times yeilds the
    # same predictions 100 times.
    #
    # Github runners indicate that they run on microsoft Azure with DS2-v2.
    # https://docs.github.com/en/actions/using-github-hosted-runners/about-github-hosted-runners#cloud-hosts-for-github-hosted-runners
    #
    # These seem to have consistent CPU's so I'm unsure what the underlying reason
    # for this to randomly fail only sometimes on Github runners
    __test__ = True

    res = dict()
    res["default_iris"] = 0.98
    res["iris_n_calls"] = 7
    res["iris_iterative_n_iter"] = 94
    res["default_iris_iterative"] = res["default_iris"]
    res["default_iris_proba"] = 0.647786774635315
    res["default_iris_proba_places"] = 6
    res["default_iris_sparse"] = 0.42
    res["default_digits"] = 0.8099574984820886
    res["digits_n_calls"] = 7
    res["digits_iterative_n_iter"] = 124
    res["default_digits_iterative"] = res["default_digits"]
    res["default_digits_binary"] = 0.99210686095932
    res["default_digits_multilabel"] = 0.8083000396415946
    res["default_digits_multilabel_proba"] = 0.8096624850657109

    sk_mod = sklearn.neural_network.MLPClassifier
    module = MLPClassifier
    step_hyperparameter = {
        "name": "n_iter_",
        "value": module.get_max_iter(),
    }



from autosklearn.pipeline.components.feature_preprocessing.kitchen_sinks import (
    RandomKitchenSinks,
)
from autosklearn.pipeline.util import PreprocessingTestCase, _test_preprocessing

import unittest


class KitchenSinkComponent(PreprocessingTestCase):
    def test_default_configuration(self):
        transformation, original = _test_preprocessing(RandomKitchenSinks)
        self.assertEqual(transformation.shape[0], original.shape[0])
        self.assertEqual(transformation.shape[1], 100)
        self.assertFalse((transformation == 0).all())

    @unittest.skip("Right now, the RBFSampler returns a float64 array!")
    def test_preprocessing_dtype(self):
        super(KitchenSinkComponent, self)._test_preprocessing_dtype(RandomKitchenSinks)



from typing import Optional

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import UniformFloatHyperparameter

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnRegressionAlgorithm
from autosklearn.pipeline.constants import DENSE, PREDICTIONS, UNSIGNED_DATA


class GaussianProcess(AutoSklearnRegressionAlgorithm):
    def __init__(self, alpha, thetaL, thetaU, random_state=None):
        self.alpha = alpha
        self.thetaL = thetaL
        self.thetaU = thetaU
        self.random_state = random_state
        self.estimator = None

    def fit(self, X, y):
        import sklearn.gaussian_process

        self.alpha = float(self.alpha)
        self.thetaL = float(self.thetaL)
        self.thetaU = float(self.thetaU)

        n_features = X.shape[1]
        kernel = sklearn.gaussian_process.kernels.RBF(
            length_scale=[1.0] * n_features,
            length_scale_bounds=[(self.thetaL, self.thetaU)] * n_features,
        )

        # Instanciate a Gaussian Process model
        self.estimator = sklearn.gaussian_process.GaussianProcessRegressor(
            kernel=kernel,
            n_restarts_optimizer=10,
            optimizer="fmin_l_bfgs_b",
            alpha=self.alpha,
            copy_X_train=True,
            random_state=self.random_state,
            normalize_y=True,
        )

        if y.ndim == 2 and y.shape[1] == 1:
            y = y.flatten()

        self.estimator.fit(X, y)

        return self

    def predict(self, X):
        if self.estimator is None:
            raise NotImplementedError
        return self.estimator.predict(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "GP",
            "name": "Gaussian Process",
            "handles_regression": True,
            "handles_classification": False,
            "handles_multiclass": False,
            "handles_multilabel": False,
            "handles_multioutput": True,
            "is_deterministic": True,
            "input": (DENSE, UNSIGNED_DATA),
            "output": (PREDICTIONS,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        alpha = UniformFloatHyperparameter(
            name="alpha", lower=1e-14, upper=1.0, default_value=1e-8, log=True
        )
        thetaL = UniformFloatHyperparameter(
            name="thetaL", lower=1e-10, upper=1e-3, default_value=1e-6, log=True
        )
        thetaU = UniformFloatHyperparameter(
            name="thetaU", lower=1.0, upper=100000, default_value=100000.0, log=True
        )

        cs = ConfigurationSpace()
        cs.add_hyperparameters([alpha, thetaL, thetaU])
        return cs



import sklearn.tree

from autosklearn.pipeline.components.classification.decision_tree import DecisionTree

from .test_base import BaseClassificationComponentTest


class DecisionTreeComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.62
    res["default_iris_iterative"] = -1
    res["default_iris_proba"] = 0.51333963481747835
    res["default_iris_sparse"] = 0.41999999999999998
    res["default_digits"] = 0.15057680631451123
    res["default_digits_iterative"] = -1
    res["default_digits_binary"] = 0.92167577413479052
    res["default_digits_multilabel"] = 0.076521739130434779
    res["default_digits_multilabel_proba"] = 0.80426747311827962

    sk_mod = sklearn.tree.DecisionTreeClassifier
    module = DecisionTree



# -*- encoding: utf-8 -*-
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import multiprocessing

import numpy as np
from ConfigSpace import Configuration
from smac.tae import StatusType

from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.evaluation.abstract_evaluator import (
    AbstractEvaluator,
    _fit_and_suppress_warnings,
)
from autosklearn.metrics import Scorer, calculate_losses
from autosklearn.pipeline.components.base import ThirdPartyComponents

__all__ = ["eval_t", "TestEvaluator"]


class TestEvaluator(AbstractEvaluator):
    def __init__(
        self,
        backend: Backend,
        queue: multiprocessing.Queue,
        metrics: Sequence[Scorer],
        additional_components: Dict[str, ThirdPartyComponents],
        port: Optional[int],
        configuration: Optional[Union[int, Configuration]] = None,
        scoring_functions: Optional[List[Scorer]] = None,
        seed: int = 1,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        disable_file_output: bool = False,
        init_params: Optional[Dict[str, Any]] = None,
    ):
        super(TestEvaluator, self).__init__(
            backend=backend,
            queue=queue,
            port=port,
            configuration=configuration,
            metrics=metrics,
            additional_components=additional_components,
            scoring_functions=scoring_functions,
            seed=seed,
            output_y_hat_optimization=False,
            num_run=-1,
            include=include,
            exclude=exclude,
            disable_file_output=disable_file_output,
            init_params=init_params,
        )
        self.configuration = configuration

        self.X_train = self.datamanager.data["X_train"]
        self.Y_train = self.datamanager.data["Y_train"]

        self.X_test = self.datamanager.data.get("X_test")
        self.Y_test = self.datamanager.data.get("Y_test")

        self.model = self._get_model(self.feat_type)

    def fit_predict_and_loss(self) -> None:
        _fit_and_suppress_warnings(self.logger, self.model, self.X_train, self.Y_train)
        loss, Y_pred, _, _ = self.predict_and_loss()
        self.finish_up(
            loss=loss,
            train_loss=None,
            opt_pred=Y_pred,
            test_pred=None,
            file_output=False,
            final_call=True,
            additional_run_info=None,
            status=StatusType.SUCCESS,
        )

    def predict_and_loss(
        self, train: bool = False
    ) -> Tuple[Union[Dict[str, float], float], np.array, Any, Any]:
        if train:
            Y_pred = self.predict_function(
                self.X_train, self.model, self.task_type, self.Y_train
            )
            err = calculate_losses(
                solution=self.Y_train,
                prediction=Y_pred,
                task_type=self.task_type,
                metrics=self.metrics,
                scoring_functions=self.scoring_functions,
            )
        else:
            Y_pred = self.predict_function(
                self.X_test, self.model, self.task_type, self.Y_train
            )
            err = calculate_losses(
                solution=self.Y_test,
                prediction=Y_pred,
                task_type=self.task_type,
                metrics=self.metrics,
                scoring_functions=self.scoring_functions,
            )

        return err, Y_pred, None, None


# create closure for evaluating an algorithm
# Has a stupid name so pytest doesn't regard it as a test
def eval_t(
    queue: multiprocessing.Queue,
    config: Union[int, Configuration],
    backend: Backend,
    metrics: Sequence[Scorer],
    seed: int,
    num_run: int,
    instance: Dict[str, Any],
    scoring_functions: Optional[List[Scorer]],
    output_y_hat_optimization: bool,
    include: Optional[List[str]],
    exclude: Optional[List[str]],
    disable_file_output: bool,
    port: Optional[int],
    additional_components: Dict[str, ThirdPartyComponents],
    init_params: Optional[Dict[str, Any]] = None,
    budget: Optional[float] = None,
    budget_type: Optional[str] = None,
) -> None:
    evaluator = TestEvaluator(
        configuration=config,
        backend=backend,
        metrics=metrics,
        seed=seed,
        port=port,
        queue=queue,
        scoring_functions=scoring_functions,
        include=include,
        exclude=exclude,
        disable_file_output=disable_file_output,
        additional_components=additional_components,
        init_params=init_params,
    )

    evaluator.fit_predict_and_loss()



# -*- encoding: utf-8 -*-

BINARY_CLASSIFICATION = 1
MULTICLASS_CLASSIFICATION = 2
MULTILABEL_CLASSIFICATION = 3
REGRESSION = 4
MULTIOUTPUT_REGRESSION = 5

REGRESSION_TASKS = [REGRESSION, MULTIOUTPUT_REGRESSION]
CLASSIFICATION_TASKS = [
    BINARY_CLASSIFICATION,
    MULTICLASS_CLASSIFICATION,
    MULTILABEL_CLASSIFICATION,
]

TASK_TYPES = REGRESSION_TASKS + CLASSIFICATION_TASKS

TASK_TYPES_TO_STRING = {
    BINARY_CLASSIFICATION: "binary.classification",
    MULTICLASS_CLASSIFICATION: "multiclass.classification",
    MULTILABEL_CLASSIFICATION: "multilabel.classification",
    REGRESSION: "regression",
    MULTIOUTPUT_REGRESSION: "multioutput.regression",
}

STRING_TO_TASK_TYPES = {
    "binary.classification": BINARY_CLASSIFICATION,
    "multiclass.classification": MULTICLASS_CLASSIFICATION,
    "multilabel.classification": MULTILABEL_CLASSIFICATION,
    "regression": REGRESSION,
    "multioutput.regression": MULTIOUTPUT_REGRESSION,
}
